
""" Implementations of the (augmented) Chebyshev scalarization technique.

This module contains implementations of the AcquisitionFunction ABC, which
use the weighted-sum technique.

The classes include:
 * ``UniformAugChebyshev`` (sample Chebyshev weights uniformly)
 * ``FixedAugChebyshev`` (uses a fixed weights, which can be set upon init)

"""

from jax import numpy as jnp
import numpy as np
import inspect
from parmoo.structs import AcquisitionFunction
from parmoo.util import xerror


class UniformAugChebyshev(AcquisitionFunction):
    """ Randomly generate weights and scalarize via augmented Chebyshev.

    Generates uniformly distributed scalarization weights, by randomly
    sampling the probability simplex.

    """

    # Slots for the UniformAugChebyshev class
    __slots__ = ['n', 'o', 'lb', 'ub', 'weights', 'alpha', 'np_rng', 'eps']

    def __init__(self, o, lb, ub, hyperparams):
        """ Constructor for the UniformAugChebyshev class.

        Args:
            o (int): The number of objectives.

            lb (numpy.ndarray): A 1d array of lower bounds for the design
                region. The number of design variables is inferred from the
                dimension of lb.

            ub (numpy.ndarray): A 1d array of upper bounds for the design
                region. The dimension must match ub.

            hyperparams (dict): A dictionary of hyperparameters for tuning
                the acquisition function. May include:
                 * 'alpha' (float): The weight to place on the linear
                   term. When not present, defaults to
                   1e-4 / number of objectives.

        Returns:
            UniformAugChebyshev: A new UniformAugChebyshev generator.

        """

        # Check inputs
        xerror(o=o, lb=lb, ub=ub, hyperparams=hyperparams)
        # Set the objective count
        self.o = o
        # Set the design variable count
        self.n = np.size(lb)
        # Set the bound constraints
        self.lb = lb
        self.ub = ub
        # Initialize the weights array
        self.weights = np.zeros(o)
        # Check hyperparameters
        self.eps = jnp.finfo(jnp.ones(1)).eps
        self.alpha = self.eps ** 0.25 / self.o
        if 'alpha' in hyperparams.keys():
            if isinstance(hyperparams['alpha'], float):
                if hyperparams['alpha'] >= 0 and hyperparams['alpha'] <= 1:
                    self.alpha = hyperparams['alpha']
                else:
                    raise ValueError("When present, hyperparams['alpha'] " +
                                     "must be in the range [0, 1]")
            else:
                raise TypeError("When present, hyperparams['alpha'] " +
                                "must be a float type")
        # Check the hyperparameter dictionary for random generator
        if 'np_random_gen' in hyperparams:
            if isinstance(hyperparams['np_random_gen'], np.random.Generator):
                self.np_rng = hyperparams['np_random_gen']
            else:
                raise TypeError("When present, hyperparams['np_random_gen'] "
                                "must be an instance of the class "
                                "numpy.random.Generator")
        else:
            self.np_rng = np.random.default_rng()
        return

    def useSD(self):
        """ Query whether this method uses uncertainties.

        When False, allows users to shortcut expensive uncertainty
        computations.

        """

        return False

    def setTarget(self, data, penalty_func):
        """ Randomly generate a new vector of scalarizing weights.

        Args:
            data (dict): A dictionary specifying the current function
                evaluation database.

            penalty_func (function): A function of one (x) or two (x, sx)
                inputs that evaluates the (penalized) objectives.

        Returns:
            numpy.ndarray: A 1d array containing the 'best' feasible starting
            point for the scalarized problem (if any previous evaluations
            were feasible) or the point in the existing database that is
            most nearly feasible.

        """

        from parmoo.util import updatePF

        # Check whether any data was given
        no_data = False
        # Check for illegal input from data
        if not isinstance(data, dict):
            raise TypeError("data must be a dict")
        else:
            if ('x_vals' in data) != ('f_vals' in data):
                raise AttributeError("if x_vals is a key in data, then " +
                                     "f_vals must also appear")
            elif 'x_vals' in data:
                if data['x_vals'] is not None and data['f_vals'] is not None:
                    if data['x_vals'].shape[0] != data['f_vals'].shape[0]:
                        raise ValueError("x_vals and f_vals must be equal " +
                                         "length")
                    if data['x_vals'].shape[1] != self.n:
                        raise ValueError("The rows of x_vals must have " +
                                         "length n")
                    if data['f_vals'].shape[1] != self.o:
                        raise ValueError("The rows of f_vals must have " +
                                         "length o")
                else:
                    no_data = True
            else:
                no_data = True
        # Check whether penalty_func() has an appropriate signature
        if callable(penalty_func):
            if len(inspect.signature(penalty_func).parameters) not in [1, 2]:
                raise ValueError("penalty_func() must accept exactly one"
                                 + " input")
        else:
            raise TypeError("penalty_func() must be callable")
        if no_data:
            # If data is empty, then the Pareto front is empty
            pf = {'x_vals': np.zeros((0, self.n)),
                  'f_vals': np.zeros((0, self.o)),
                  'c_vals': np.zeros((0, 1))}
        else:
            # Get the Pareto front
            pf = updatePF(data, {})
        # Sample the weights uniformly from the unit simplex
        self.weights = -np.log(1.0 - self.np_rng.random(self.o))
        self.weights = self.weights[:] / sum(self.weights[:])
        # If data is empty, randomly select weights and starting point
        if no_data:
            # Randomly select a starting point
            x_start = (self.np_rng.random(self.n) * (self.ub - self.lb)
                       + self.lb)
            return x_start
        # If data is nonempty but pf is empty, use a penalty to select
        elif pf is None or pf['x_vals'].shape[0] == 0:
            x_best = np.zeros(data['x_vals'].shape[1])
            p_best = np.inf
            for xi, fi, ci in zip(data['x_vals'], data['f_vals'],
                                  data['c_vals']):
                p_temp = np.sum(fi) / np.sqrt(self.eps) + np.sum(ci)
                if p_temp < p_best:
                    x_best = xi
                    p_best = p_temp
            return x_best
        else:
            xx = np.zeros(1)
            sx = np.zeros(1)
            sdx = np.zeros(1)
            i = np.argmin(np.asarray([self.scalarize(fi, xx, sx, sdx)
                                      for fi in pf['f_vals']]))
            x = pf['x_vals'][i, :]
            return x

    def scalarize(self, f_vals, x_vals, s_vals_mean, s_vals_sd, manifold=None):
        """ Scalarize a vector of function values using the current weights.

        Args:
            f_vals (numpy.ndarray): A 1d array specifying the function
                values to be scalarized.

            x_vals (np.ndarray): A 1D array specifying a vector the design
                point corresponding to f_vals (unused by this method).

            s_vals_mean (np.ndarray): A 1D array specifying the expected
                simulation outputs for the x value being scalarized
                (unused by this method).

            s_vals_sd (np.ndarray): A 1D array specifying the standard
                deviation for each of the simulation outputs (unused by
                this method).

        Returns:
            float: The scalarized value.

        """

        if not isinstance(manifold, int):
            return (jnp.max(f_vals * self.weights)
                    + self.alpha * jnp.sum(f_vals))
        else:
            return ((f_vals * self.weights)[manifold]
                    + self.alpha * jnp.sum(f_vals))

    def getManifold(self, f_vals):
        """ Check which manifold is active for a given function value.

        Each component of f_vals is its own smooth manifold, but once
        the max function is applied for the augmented Chebyshev scalarization,
        only 1 component is typically active in the surrogate
        minimization problem.

        Args:
            f_vals (numpy.ndarray): A 1d array specifying the function
                values to be scalarized.

        Returns:
            numpy.ndarray, dtype=int: A 1d array of integers, matching the
            length of f_vals, where
             - 0 indicates that this component of the manifold is inactive and
             - 1 indicates that this component of the manifold is active.

        """

        return np.isclose(f_vals * self.weights,
                          (f_vals * self.weights).max(),
                          atol=np.sqrt(self.eps)).astype(int)


class FixedAugChebyshev(AcquisitionFunction):
    """ Use fixed weights with the augmented Chebyshev scalarization.

    Use a fixed scalarization scheme, based on a fixed weighted sum.

    """

    # Slots for the FixedAugChebyshev class
    __slots__ = ['n', 'o', 'lb', 'ub', 'weights', 'alpha', 'np_rng', 'eps']

    def __init__(self, o, lb, ub, hyperparams):
        """ Constructor for the FixedAugChebyshev class.

        Args:
            o (int): The number of objectives.

            lb (numpy.ndarray): A 1d array of lower bounds for the design
                region. The number of design variables is inferred from the
                dimension of lb.

            ub (numpy.ndarray): A 1d array of upper bounds for the design
                region. The dimension must match ub.

            hyperparams (dict): A dictionary of hyperparameters for tuning
                the acquisition function. May contain the following key:
                 * 'weights' (numpy.ndarray): A 1d array of length o that,
                   when present, specifies the scalarization weights to use.
                   When absent, the default weights are w = [1/o, ..., 1/o].
                 * 'alpha' (float): The weight to place on the linear
                   term. When not present, defaults to
                   1e-4 / number of objectives.

        Returns:
            FixedAugChebyshev: A new FixedAugChebyshev generator.

        """

        # Check inputs
        xerror(o=o, lb=lb, ub=ub, hyperparams=hyperparams)
        # Set the objective count
        self.o = o
        # Set the design variable count
        self.n = np.size(lb)
        # Set the bound constraints
        self.lb = lb
        self.ub = ub
        # Check the hyperparameter dictionary for random generator
        if 'np_random_gen' in hyperparams:
            if isinstance(hyperparams['np_random_gen'], np.random.Generator):
                self.np_rng = hyperparams['np_random_gen']
            else:
                raise TypeError("When present, hyperparams['np_random_gen'] "
                                "must be an instance of the class "
                                "numpy.random.Generator")
        else:
            self.np_rng = np.random.default_rng()
        # Check the hyperparams dictionary for weights
        if 'weights' in hyperparams:
            # If weights are provided, check that they are legal
            if not isinstance(hyperparams['weights'], np.ndarray):
                raise TypeError("when present, 'weights' must be a " +
                                "numpy array")
            else:
                if hyperparams['weights'].size != self.o:
                    raise ValueError("when present, 'weights' must " +
                                     "have length o")
                else:
                    # Assign the weights
                    self.weights = hyperparams['weights'].flatten()
        else:
            # If no weights provided, sample from the unit simplex
            self.weights = -np.log(1.0 - self.np_rng.random(self.o))
            self.weights = self.weights[:] / sum(self.weights[:])
        # Check hyperparameters dictionary for alpha
        self.eps = jnp.finfo(jnp.ones(1)).eps
        self.alpha = self.eps ** 0.25 / self.o
        if 'alpha' in hyperparams.keys():
            if isinstance(hyperparams['alpha'], float):
                if hyperparams['alpha'] >= 0 and hyperparams['alpha'] <= 1:
                    self.alpha = hyperparams['alpha']
                else:
                    raise ValueError("When present, hyperparams['alpha'] " +
                                     "must be in the range [0, 1]")
            else:
                raise TypeError("When present, hyperparams['alpha'] " +
                                "must be a float type")
        return

    def useSD(self):
        """ Query whether this method uses uncertainties.

        When False, allows users to shortcut expensive uncertainty
        computations.

        """

        return False

    def setTarget(self, data, penalty_func):
        """ Randomly generate a feasible starting point.

        Args:
            data (dict): A dictionary specifying the current function
                evaluation database.

            penalty_func (function): A function of one (x) or two (x, sx)
                inputs that evaluates the (penalized) objectives.

        Returns:
            numpy.ndarray: A 1d array containing the 'best' feasible starting
            point for the scalarized problem (if any previous evaluations
            were feasible) or the point in the existing database that is
            most nearly feasible.

        """

        from parmoo.util import updatePF

        # Check whether any data was given
        no_data = False
        # Check for illegal input from data
        if not isinstance(data, dict):
            raise TypeError("data must be a dict")
        else:
            if ('x_vals' in data) != ('f_vals' in data):
                raise AttributeError("if x_vals is a key in data, then " +
                                     "f_vals must also appear")
            elif 'x_vals' in data:
                if data['x_vals'] is not None and data['f_vals'] is not None:
                    if data['x_vals'].shape[0] != data['f_vals'].shape[0]:
                        raise ValueError("x_vals and f_vals must be equal " +
                                         "length")
                    if data['x_vals'].shape[1] != self.n:
                        raise ValueError("The rows of x_vals must have " +
                                         "length n")
                    if data['f_vals'].shape[1] != self.o:
                        raise ValueError("The rows of f_vals must have " +
                                         "length o")
                else:
                    no_data = True
            else:
                no_data = True
        # Check whether penalty_func() has an appropriate signature
        if callable(penalty_func):
            if len(inspect.signature(penalty_func).parameters) not in [1, 2]:
                raise ValueError("penalty_func() must accept exactly one"
                                 + " input")
        else:
            raise TypeError("penalty_func() must be callable")
        if no_data:
            # If data is empty, then the Pareto front is empty
            pf = {'x_vals': np.zeros((0, self.n)),
                  'f_vals': np.zeros((0, self.o)),
                  'c_vals': np.zeros((0, 1))}
        else:
            # Get the Pareto front
            pf = updatePF(data, {})
        # If data is empty, randomly select weights and starting point
        if no_data:
            # Randomly select a starting point
            x_start = (self.np_rng.random(self.n) * (self.ub - self.lb)
                       + self.lb)
            return x_start
        # If data is nonempty but pf is empty, use a penalty to select
        elif pf is None or pf['x_vals'].shape[0] == 0:
            x_best = np.zeros(data['x_vals'].shape[1])
            p_best = np.inf
            for xi, fi, ci in zip(data['x_vals'], data['f_vals'],
                                  data['c_vals']):
                p_temp = np.sum(fi) / np.sqrt(self.eps) + np.sum(ci)
                if p_temp < p_best:
                    x_best = xi
                    p_best = p_temp
            return x_best
        else:
            xx = np.zeros(1)
            sx = np.zeros(1)
            sdx = np.zeros(1)
            i = np.argmin(np.asarray([self.scalarize(fi, xx, sx, sdx)
                                      for fi in pf['f_vals']]))
            x = pf['x_vals'][i, :]
            return x

    def scalarize(self, f_vals, x_vals, s_vals_mean, s_vals_sd, manifold=None):
        """ Scalarize a vector of function values using the current weights.

        Args:
            f_vals (numpy.ndarray): A 1d array specifying the function
                values to be scalarized.

            x_vals (np.ndarray): A 1D array specifying a vector the design
                point corresponding to f_vals (unused by this method).

            s_vals_mean (np.ndarray): A 1D array specifying the expected
                simulation outputs for the x value being scalarized
                (unused by this method).

            s_vals_sd (np.ndarray): A 1D array specifying the standard
                deviation for each of the simulation outputs (unused by
                this method).

        Returns:
            float: The scalarized value.

        """

        if not isinstance(manifold, int):
            return (jnp.max(f_vals * self.weights)
                    + self.alpha * jnp.sum(f_vals))
        else:
            return ((f_vals * self.weights)[manifold]
                    + self.alpha * jnp.sum(f_vals))

    def getManifold(self, f_vals):
        """ Check which manifold is active for a given function value.

        Each component of f_vals is its own smooth manifold, but once
        the max function is applied for the augmented Chebyshev scalarization,
        only 1 component is typically active in the surrogate
        minimization problem.

        Args:
            f_vals (numpy.ndarray): A 1d array specifying the function
                values to be scalarized.

        Returns:
            numpy.ndarray, dtype=int: A 1d array of integers, matching the
            length of f_vals, where
             - 0 indicates that this component of the manifold is inactive and
             - 1 indicates that this component of the manifold is active.

        """

        return np.isclose(f_vals * self.weights,
                          (f_vals * self.weights).max(),
                          atol=np.sqrt(self.eps)).astype(int)




def test_MOOP_init():
    """ Check that the MOOP class handles initialization properly.

    Initialize several MOOP objects, and check that their internal fields
    appear correct.

    """

    from parmoo import MOOP
    from parmoo.optimizers import LocalSurrogate_PS
    import pytest

    # Try several invalid inputs
    with pytest.raises(TypeError):
        MOOP(5.0)
    with pytest.raises(TypeError):
        MOOP(lambda w, x, y, z: 0.0)
    with pytest.raises(TypeError):
        MOOP(LocalSurrogate_PS, hyperparams=[])
    # Try a few valid inputs
    moop = MOOP(LocalSurrogate_PS)
    assert (moop.m == 0 and moop.n_feature == 0 and moop.n_latent == 0 and
            moop.s == 0 and moop.o == 0 and moop.p == 0)
    moop = MOOP(LocalSurrogate_PS, hyperparams={'test': 0})
    assert (moop.m == 0 and moop.n_feature == 0 and moop.n_latent == 0 and
            moop.s == 0 and moop.o == 0 and moop.p == 0)
    assert (moop.opt_hp['test'] == 0)


def test_MOOP_addDesign():
    """ Check that the MOOP class handles adding design variables properly.

    Initialize a MOOP objects, and add several design variables.

    """

    from parmoo import MOOP
    from parmoo.embeddings import IdentityEmbedder
    from parmoo.optimizers import LocalSurrogate_PS
    import pytest

    # Initialize a MOOP with no hyperparameters
    moop = MOOP(LocalSurrogate_PS)
    # Try to add some bad design variable types
    with pytest.raises(TypeError):
        moop.addDesign([])
    with pytest.raises(TypeError):
        moop.addDesign({'des_type': 1.0})
    with pytest.raises(ValueError):
        moop.addDesign({'des_type': "hello world"})
    assert (moop.n_latent == 0)
    # Now add some continuous and integer design variables
    moop.addDesign({'lb': 0.0,
                    'ub': 1.0})
    # Try to use a repeated name to test error handling
    with pytest.raises(ValueError):
        moop.addDesign({'name': "x1", 'lb': 0.0, 'ub': 1.0})
    assert (moop.n_latent == 1)
    moop.addDesign({'name': "x2",
                    'des_type': "continuous",
                    'lb': 0.0,
                    'ub': 1.0,
                    'des_tol': 0.01})
    assert (moop.n_latent == 2)
    moop.addDesign({'des_type': "integer",
                    'lb': 0,
                    'ub': 4})
    assert (moop.n_latent == 3)
    # Now add some categorical design variables
    moop.addDesign({'des_type': "categorical",
                    'levels': 2})
    assert (moop.n_latent == 4)
    moop.addDesign({'des_type': "categorical",
                    'levels': 3})
    assert (moop.n_latent == 6)
    moop.addDesign({'name': "x6",
                    'des_type': "categorical",
                    'levels': ["boy", "girl", "doggo"]})
    assert (moop.n_latent == 8)
    # Now add a custom design variables
    moop.addDesign({'des_type': "custom",
                    'lb': -100.0,
                    'ub': 100.0,
                    'embedder': IdentityEmbedder})
    assert (moop.n_latent == 9)
    moop.addDesign({'des_type': "raw",
                    'lb': -100.0,
                    'ub': 100.0})
    assert (moop.n_latent == 10)


def test_MOOP_addSimulation():
    """ Check that the MOOP class handles adding new simulations properly.

    Initialize several MOOPs, and add several simulations. Check that
    the metadata is updated correctly.

    """

    import numpy as np
    from parmoo import MOOP
    from parmoo.optimizers import LocalSurrogate_PS
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF
    import pytest

    # Initialize a MOOP and add 3 design variables
    moop = MOOP(LocalSurrogate_PS)
    moop.addDesign({'des_type': "continuous",
                    'lb': 0.0,
                    'ub': 1.0})
    moop.addDesign({'des_type': "continuous",
                    'lb': 0.0,
                    'ub': 1.0})
    moop.addDesign({'des_type': "continuous",
                    'lb': 0.0,
                    'ub': 1.0})
    # Now add one simulation and check
    g1 = {'m': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm(x)],
          'surrogate': GaussRBF}
    moop.addSimulation(g1)
    assert (moop.m == 1 and moop.n_latent == 3 and moop.s == 1 and
            moop.o == 0 and moop.p == 0)
    # Initialize another MOOP with 3 design variables
    moop = MOOP(LocalSurrogate_PS)
    moop.addDesign({'des_type': "continuous",
                    'lb': 0.0,
                    'ub': 1.0})
    moop.addDesign({'des_type': "continuous",
                    'lb': 0.0,
                    'ub': 1.0})
    moop.addDesign({'des_type': "continuous",
                    'lb': 0.0,
                    'ub': 1.0})
    g2 = {'m': 2,
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm(x-1.0), np.linalg.norm(x-0.5)],
          'surrogate': GaussRBF}
    moop.addSimulation(g1, g2)
    assert (moop.m == 3 and moop.n_latent == 3 and moop.s == 2 and
            moop.o == 0 and moop.p == 0)
    g3 = {'name': "Bobo1",
          'm': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm(x)],
          'surrogate': GaussRBF}
    g4 = {'name': "Bobo2",
          'm': 2,
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm(x-1.0), np.linalg.norm(x-0.5)],
          'surrogate': GaussRBF}
    moop.addSimulation(g3, g4)
    # Try to use a repeated name to test error handling
    with pytest.raises(ValueError):
        moop.addSimulation(g4)
    # Check the names
    assert (moop.sim_schema[0][0] == "sim1")
    assert (moop.sim_schema[1][0] == "sim2")
    assert (moop.sim_schema[2][0] == "Bobo1")
    assert (moop.sim_schema[3][0] == "Bobo2")


def test_MOOP_addObjective():
    """ Check that the MOOP class handles adding objectives properly.

    Initialize a MOOP object and check that the addObjective() function works
    correctly.

    """

    from parmoo import MOOP
    from parmoo.optimizers import LocalSurrogate_PS
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF
    import pytest

    # Initialize a MOOP with 2 SimGroups, one of which has 2 outputs
    moop = MOOP(LocalSurrogate_PS)
    for i in range(3):
        moop.addDesign({'lb': 0.0, 'ub': 1.0})
    g1 = {'n': 3,
          'm': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: 0.0,
          'surrogate': GaussRBF}
    g2 = {'n': 3,
          'm': 2,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [1.0, 0.5],
          'surrogate': GaussRBF}
    moop.addSimulation(g1, g2)
    # Try to add bad objectives and check that appropriate errors are raised
    with pytest.raises(TypeError):
        moop.addObjective(0)
    with pytest.raises(AttributeError):
        moop.addObjective({})
    with pytest.raises(TypeError):
        moop.addObjective({'obj_func': 0})
    with pytest.raises(ValueError):
        moop.addObjective({'obj_func': lambda x: 0.0})
    # Check that no objectives were added yet
    assert (moop.o == 0)
    # Now add 3 good objectives
    moop.addObjective({'obj_func': lambda x, s: 0.0})
    assert (moop.o == 1)
    moop.addObjective({'obj_func': lambda x, s: 0.0},
                      {'obj_func': lambda x, s: 0.0,
                       'obj_grad': lambda x, s: (x, s)})
    assert (moop.o == 3)
    moop.addObjective({'name': "Bobo", 'obj_func': lambda x, s: 0.0})
    assert (moop.o == 4)
    # Try to use a repeated name to test error handling
    with pytest.raises(ValueError):
        moop.addObjective({'name': "Bobo", 'obj_func': lambda x, s: 0.0})
    assert (moop.obj_schema[0] == ("f1", 'f8'))
    assert (moop.obj_schema[1] == ("f2", 'f8'))
    assert (moop.obj_schema[2] == ("f3", 'f8'))
    assert (moop.obj_schema[3] == ("Bobo", 'f8'))


def test_MOOP_addConstraint():
    """ Check that the MOOP class handles adding constraints properly.

    Initialize a MOOP object and check that the addConstraint() function works
    correctly.

    """

    from parmoo import MOOP
    from parmoo.optimizers import LocalSurrogate_PS
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF
    import pytest

    # Initialize a MOOP with 2 SimGroups, one of which has 2 outputs
    moop = MOOP(LocalSurrogate_PS)
    for i in range(3):
        moop.addDesign({'lb': 0.0, 'ub': 1.0})
    g1 = {'n': 3,
          'm': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: 0.0,
          'surrogate': GaussRBF}
    g2 = {'n': 3,
          'm': 2,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [1.0, 0.5],
          'surrogate': GaussRBF}
    moop.addSimulation(g1, g2)
    # Try to add bad constraints and check that appropriate errors are raised
    with pytest.raises(TypeError):
        moop.addConstraint(0)
    with pytest.raises(AttributeError):
        moop.addConstraint({})
    with pytest.raises(TypeError):
        moop.addConstraint({'con_func': 0})
    with pytest.raises(ValueError):
        moop.addConstraint({'con_func': lambda x: 0.0})
    with pytest.raises(TypeError):
        moop.addConstraint({'constraint': 0})
    with pytest.raises(ValueError):
        moop.addConstraint({'constraint': lambda x: 0.0})
    # Check that no constraints were added yet
    assert (moop.p == 0)
    # Now add 3 good constraints
    moop.addConstraint({'constraint': lambda x, s: 0.0})
    assert (moop.p == 1)
    moop.addConstraint({'con_func': lambda x, s: 0.0},
                       {'con_func': lambda x, s: 0.0,
                        'con_grad': lambda x, s: (x, s)})
    assert (moop.p == 3)
    moop.addConstraint({'name': "Bobo", 'constraint': lambda x, s: 0.0})
    assert (moop.p == 4)
    # Try to use a repeated name to test error handling
    with pytest.raises(ValueError):
        moop.addConstraint({'name': "Bobo", 'con_func': lambda x, s: 0.0})
    assert (moop.con_schema[0] == ("c1", 'f8'))
    assert (moop.con_schema[1] == ("c2", 'f8'))
    assert (moop.con_schema[2] == ("c3", 'f8'))
    assert (moop.con_schema[3] == ("Bobo", 'f8'))


def test_MOOP_addAcquisition():
    """ Check that the MOOP class handles adding acquisitions properly.

    Initialize a MOOP object and check that the addAcquisition() function works
    correctly.

    """

    from parmoo import MOOP
    from parmoo.acquisitions import UniformWeights
    from parmoo.optimizers import LocalSurrogate_PS
    from parmoo.surrogates import GaussRBF
    import pytest

    # Initialize a MOOP with 3 variables and 3 objectives
    moop = MOOP(LocalSurrogate_PS)
    for i in range(3):
        moop.addDesign({'lb': 0.0, 'ub': 1.0})
    moop.addObjective({'obj_func': lambda x, s: 0.0},
                      {'obj_func': lambda x, s: 0.0},
                      {'obj_func': lambda x, s: 0.0})
    # Try to add bad acquisition functions and check for an appropriate error
    with pytest.raises(TypeError):
        moop.addAcquisition(0)
    with pytest.raises(AttributeError):
        moop.addAcquisition({})
    with pytest.raises(TypeError):
        moop.addAcquisition({'acquisition': UniformWeights,
                             'hyperparams': 0})
    with pytest.raises(TypeError):
        moop.addAcquisition({'acquisition': 0,
                             'hyperparams': {}})
    with pytest.raises(TypeError):
        moop.addAcquisition({'acquisition': GaussRBF,
                             'hyperparams': {}})
    # Check that no acquisitions were added then add 3 good acquisitions
    assert (len(moop.acquisitions) == 0)
    moop.addAcquisition({'acquisition': UniformWeights})
    moop.addAcquisition({'acquisition': UniformWeights},
                        {'acquisition': UniformWeights, 'hyperparams': {}})
    moop.compile()
    assert (len(moop.acquisitions) == 3)


def test_MOOP_getTypes():
    """ Check that the MOOP class handles getting dtypes properly.

    Initialize a MOOP object, add design variables, simulations, objectives,
    and constraints, and get the corresponding types.

    """

    import numpy as np
    from parmoo import MOOP
    from parmoo.optimizers import LocalSurrogate_PS
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF

    # Create a new MOOP
    moop = MOOP(LocalSurrogate_PS)
    # Check that all types are None
    assert (moop.getDesignType() is None)
    assert (moop.getSimulationType() is None)
    assert (moop.getObjectiveType() is None)
    assert (moop.getConstraintType() is None)
    # Add some variables, simulations, objectives, and constraints
    moop = MOOP(LocalSurrogate_PS)
    moop.addDesign({'name': "x1", 'lb': 0.0, 'ub': 1.0})
    moop.addDesign({'name': "x2", 'des_type': "categorical", 'levels': 3})
    g1 = {'m': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [0.0],
          'surrogate': GaussRBF}
    moop.addSimulation(g1)
    moop.addObjective({'obj_func': lambda x, s: 0.0})
    moop.addConstraint({'constraint': lambda x, s: 0.0})
    # Check the dtypes
    assert (np.zeros(1, dtype=moop.getDesignType()).size == 1)
    assert (np.zeros(1, dtype=moop.getSimulationType()).size == 1)
    assert (np.zeros(1, dtype=moop.getObjectiveType()).size == 1)
    assert (np.zeros(1, dtype=moop.getConstraintType()).size == 1)


if __name__ == "__main__":
    test_MOOP_init()
    test_MOOP_addDesign()
    test_MOOP_addSimulation()
    test_MOOP_addObjective()
    test_MOOP_addConstraint()
    test_MOOP_addAcquisition()
    test_MOOP_getTypes()




import numpy as np
from parmoo import MOOP
from parmoo.searches import LatinHypercube
from parmoo.surrogates import GaussRBF
from parmoo.acquisitions import UniformWeights
from parmoo.optimizers import GlobalSurrogate_PS

my_moop = MOOP(GlobalSurrogate_PS)

my_moop.addDesign({'name': "x1",
                   'des_type': "continuous",
                   'lb': 0.0, 'ub': 1.0})
my_moop.addDesign({'name': "x2", 'des_type': "categorical",
                   'levels': 3})

def sim_func(x):
   if x["x2"] == 0:
      return np.array([(x["x1"] - 0.2) ** 2, (x["x1"] - 0.8) ** 2])
   else:
      return np.array([99.9, 99.9])

my_moop.addSimulation({'name': "MySim",
                       'm': 2,
                       'sim_func': sim_func,
                       'search': LatinHypercube,
                       'surrogate': GaussRBF,
                       'hyperparams': {'search_budget': 20}})

my_moop.addObjective({'name': "f1", 'obj_func': lambda x, s: s["MySim"][0]})
my_moop.addObjective({'name': "f2", 'obj_func': lambda x, s: s["MySim"][1]})

my_moop.addAcquisition({'acquisition': UniformWeights})

# This step is needed to finalize the MOOP definition. If you are using the
# solve command it is done automatically, but it must be done manually before
# any pre-existing data can be added.
my_moop.compile()

# Precompute one simulation value for demo
des_val = np.zeros(1, dtype=[("x1", float), ("x2", int)])[0]
sim_val = sim_func(des_val)

# Add the precomputed simulation value from above
my_moop.updateSimDb(des_val, sim_val, "MySim")

# Get and display initial database
sim_db = my_moop.getSimulationData()
print(sim_db)




def test_MOOP_iterate():
    """ Test the MOOP class's iterator in objectives.py.

    Initialize several MOOP objects and perform iterations to produce
    a batch of candidate solutions.

    """

    from jax import config
    config.update("jax_enable_x64", True)
    import numpy as np
    from parmoo import MOOP
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF
    from parmoo.acquisitions import UniformWeights
    from parmoo.optimizers import LocalSurrogate_PS, LocalSurrogate_BFGS
    import pytest

    # Initialize two simulation groups with 1 output each
    s1 = {'m': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm([x[f"x{key}"]
                                 for key in [1, 2, 3]])],
          'surrogate': GaussRBF,
          'search_budget': 20}
    s2 = {'m': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm([x[f"x{key}"] - 1
                                 for key in [1, 2, 3]])],
          'surrogate': GaussRBF,
          'search_budget': 20}
    # Create a MOOP with 3 design variables and 2 simulations
    moop1 = MOOP(LocalSurrogate_PS,
                 hyperparams={'opt_budget': 100, 'np_random_gen': 0})
    with pytest.raises(AttributeError):
        moop1.iterate(1)
    for i in range(3):
        moop1.addDesign({'name': f"x{i+1}", 'lb': 0.0, 'ub': 1.0})
    moop1.addSimulation(s1, s2)
    with pytest.raises(AttributeError):
        moop1.iterate(1)
    # Now add the two objectives
    def f1(x, sim): return sim["sim1"]
    def f2(x, sim): return sim["sim2"]
    moop1.addObjective({'obj_func': f1},
                       {'obj_func': f2})
    # Add 3 acquisition functions
    for i in range(3):
        moop1.addAcquisition({'acquisition': UniformWeights})
    # Try some invalid iterations
    with pytest.raises(ValueError):
        moop1.iterate(-1)
    with pytest.raises(TypeError):
        moop1.iterate(2.0)
    # Solve the MOOP with 1 iteration
    batch = moop1.iterate(0)
    batch = moop1.filterBatch(batch)
    for (x, i) in batch:
        moop1.evaluateSimulation(x, i)
    moop1.updateAll(0, batch)
    batch = moop1.iterate(1)
    batch = moop1.filterBatch(batch)
    for (x, i) in batch:
        moop1.evaluateSimulation(x, i)
    moop1.updateAll(1, batch)
    soln = moop1.getPF()
    # Assert that solutions were found
    assert (soln.size > 0)
    # Assert that the x_vals and f_vals match
    for si in soln:
        assert (np.abs(s1['sim_func'](si) - si['f1']) < 1.0e-8)
        assert (np.linalg.norm(s2['sim_func'](si) - si['f2']) < 1.0e-8)

    s3 = {'m': 4,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [x[i] for i in x],
          'surrogate': GaussRBF,
          'search_budget': 500}
    # Create a three objective toy problem, with one simulation
    moop2 = MOOP(LocalSurrogate_BFGS,
                 hyperparams={'opt_budget': 100, 'np_random_gen': 0})
    for i in range(4):
        moop2.addDesign({'lb': 0.0, 'ub': 1.0, 'des_tol': 0.1})
    moop2.addSimulation(s3)

    # Now add the three objectives
    def f3(x, sim):
        return np.linalg.norm(sim["sim1"] - 0.1 * np.eye(4)[0, :]) ** 2.0

    def g3(x, sim):
        x_out = x.copy()
        for key in x:
            x_out[key] = 0.0
        s_out = sim.copy()
        s_out["sim1"] *= 2.0
        s_out["sim1"] = s_out["sim1"].at[0].set(s_out["sim1"][0] - 0.2)
        return x_out, s_out

    def f4(x, sim):
        return np.linalg.norm(sim["sim1"] - 0.1 * np.eye(4)[1, :]) ** 2.0

    def g4(x, sim):
        x_out = x.copy()
        for key in x:
            x_out[key] = 0.0
        s_out = sim.copy()
        s_out["sim1"] *= 2.0
        s_out["sim1"] = s_out["sim1"].at[1].set(s_out["sim1"][1] - 0.2)
        return x_out, s_out

    def f5(x, sim):
        return np.linalg.norm(sim["sim1"] - 0.1 * np.eye(4)[2, :]) ** 2.0

    def g5(x, sim):
        x_out = x.copy()
        for key in x:
            x_out[key] = 0.0
        s_out = sim.copy()
        s_out["sim1"] *= 2.0
        s_out["sim1"] = s_out["sim1"].at[2].set(s_out["sim1"][2] - 0.2)
        return x_out, s_out

    moop2.addObjective({'obj_func': f3, 'obj_grad': g3},
                       {'obj_func': f4, 'obj_grad': g4},
                       {'obj_func': f5, 'obj_grad': g5})
    # Add 3 acquisition functions
    for i in range(3):
        moop2.addAcquisition({'acquisition': UniformWeights})
    # Do 2 iterates of the MOOP and extract the final database
    moop2.iterate(0)
    batch = []
    xi = {"x1": 0, "x2": 0, "x3": 0, "x4": 0}
    for i in range(1, 5):
        xi[f"x{i}"] = 0.1
        batch.append((xi.copy(), "sim1"))
        xi[f"x{i}"] = 0
    batch.append(({"x1": 0.1, "x2": 0.1, "x3": 0.1, "x4": 0.1}, "sim1"))
    for (x, i) in batch:
        moop2.evaluateSimulation(x, i)
    moop2.updateAll(0, batch)
    batch = moop2.iterate(1)
    batch = moop2.filterBatch(batch)
    for (x, i) in batch:
        moop2.evaluateSimulation(x, i)
    moop2.updateAll(1, batch)
    soln = moop2.getPF()
    # Assert that solutions were found
    assert (soln.size > 0)
    # Assert that the x_vals and f_vals match
    si = {}
    for xsi in soln:
        xi = {"x1": xsi["x1"],
              "x2": xsi["x2"],
              "x3": xsi["x3"],
              "x4": xsi["x4"]}
        si["sim1"] = s3["sim_func"](xi)
        fi = [xsi["f1"], xsi["f2"], xsi["f3"]]
        assert (np.linalg.norm(np.array([f3(xi, si), f4(xi, si), f5(xi, si)]
                                        ).flatten() - fi) < 1.0e-8)
        for j in xi:
            assert (xi[j] <= 0.2)

    s4 = {'m': 4,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func':
          lambda x: abs(x["x5"] - 1) + np.array([x[f"x{i+1}"]
                                                 for i in range(4)]),
          'surrogate': GaussRBF,
          'search_budget': 500}
    # Create a three objective toy problem, with one simulation
    moop3 = MOOP(LocalSurrogate_BFGS, hyperparams={'np_random_gen': 0})
    for i in range(4):
        moop3.addDesign({'lb': 0.0, 'ub': 1.0})
    moop3.addDesign({'des_type': "categorical", 'levels': 3})
    moop3.addSimulation(s4)

    # Now add the three objectives
    def f6(x, sim):
        return np.linalg.norm(sim["sim1"] - np.eye(4)[0, :]) ** 2

    def g6(x, sim):
        dx = {"x1": 0, "x2": 0, "x3": 0, "x4": 0, "x5": 0}
        ds = {"sim1": np.array([2.0 * sim["sim1"][0] - 2.0,
                                2.0 * sim["sim1"][1],
                                2.0 * sim["sim1"][2],
                                2.0 * sim["sim1"][3]])}
        return dx, ds

    def f7(x, sim):
        return np.linalg.norm(sim["sim1"] - np.eye(4)[1, :]) ** 2

    def g7(x, sim):
        dx = {"x1": 0, "x2": 0, "x3": 0, "x4": 0, "x5": 0}
        ds = {"sim1": np.array([2.0 * sim["sim1"][0],
                                2.0 * sim["sim1"][1] - 2.0,
                                2.0 * sim["sim1"][2],
                                2.0 * sim["sim1"][3]])}
        return dx, ds

    def f8(x, sim):
        return np.linalg.norm(sim["sim1"] - np.eye(4)[2, :]) ** 2

    def g8(x, sim):
        dx = {"x1": 0, "x2": 0, "x3": 0, "x4": 0, "x5": 0}
        ds = {"sim1": np.array([2.0 * sim["sim1"][0],
                                2.0 * sim["sim1"][1],
                                2.0 * sim["sim1"][2] - 2.0,
                                2.0 * sim["sim1"][3]])}
        return dx, ds

    moop3.addObjective({'obj_func': f6, 'obj_grad': g6},
                       {'obj_func': f7, 'obj_grad': g7},
                       {'obj_func': f8, 'obj_grad': g8})
    # Add 3 acquisition functions
    for i in range(3):
        moop3.addAcquisition({'acquisition': UniformWeights})
    # Do 2 iterates of the MOOP and extract the final database
    moop3.iterate(0)
    batch = []
    for i in range(1, 6):
        xi = {"x1": 0, "x2": 0, "x3": 0, "x4": 0, "x5": 0}
        xi[f"x{i}"] = 1
        batch.append((xi.copy(), "sim1"))
    batch.append(({"x1": 1, "x2": 1, "x3": 1, "x4": 1, "x5": 1}, "sim1"))
    batch.append(({"x1": 1, "x2": 1, "x3": 1, "x4": 1, "x5": 2}, "sim1"))
    for (x, i) in batch:
        moop3.evaluateSimulation(x, i)
    moop3.updateAll(0, batch)
    batch = moop3.iterate(1)
    batch = moop3.filterBatch(batch)
    for (x, i) in batch:
        moop3.evaluateSimulation(x, i)
    moop3.updateAll(1, batch)
    soln = moop3.getPF()
    # Assert that solutions were found
    assert (np.size(soln) > 0)
    # Assert that the x_vals and f_vals match
    si = {}
    for xsi in soln:
        xi = {"x1": xsi["x1"],
              "x2": xsi["x2"],
              "x3": xsi["x3"],
              "x4": xsi["x4"],
              "x5": xsi["x5"]}
        si["sim1"] = s4["sim_func"](xi)
        fi = [xsi["f1"], xsi["f2"], xsi["f3"]]
        assert (np.linalg.norm(np.array([f6(xi, si), f7(xi, si), f8(xi, si)]
                                        ).flatten() - fi) < 1.0e-8)
        assert (abs(xi["x4"]) <= 0.1 and abs(xi["x5"] - 1) <= 0.1)

    s5 = {'m': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [(x["x0"] - 1.0) * (x["x0"] - 1.0) +
                                 (x["x1"]) * (x["x1"]) + float(x["x2"])],
          'surrogate': GaussRBF,
          'search_budget': 100}
    # Solve a MOOP with categorical variables
    moop4 = MOOP(LocalSurrogate_BFGS, hyperparams={'np_random_gen': 0})
    moop4.addDesign({'name': "x0", 'lb': 0.0, 'ub': 1.0})
    moop4.addDesign({'name': "x1", 'lb': 0.0, 'ub': 1.0})
    moop4.addDesign({'name': "x2", 'des_type': "categorical",
                     'levels': ["0", "1"]})
    moop4.addSimulation(s5)

    # Now add the two objectives
    def f9(x, sim):
        return sim["sim1"]

    def g9(x, sim):
        dx = {"x0": 0, "x1": 0}
        ds = {"sim1": 1}
        return dx, ds

    def f10(x, sim):
        return ((x["x0"]) * (x["x0"]) +
                (x["x1"] - 1.0) * (x["x1"] - 1.0) + float(x["x2"]))

    def g10(x, sim):
        dx = {}
        dx['x0'] = 2.0 * x["x0"]
        dx['x1'] = 2.0 * x["x1"] - 2.0
        ds = {"sim1": 0}
        return dx, ds

    moop4.addObjective({'obj_func': f9, 'obj_grad': g9},
                       {'obj_func': f10, 'obj_grad': g10})
    # Add 3 acquisition functions
    for i in range(3):
        moop4.addAcquisition({'acquisition': UniformWeights})
    # Do 2 iterates of the MOOP and extract the final database
    batch = moop4.iterate(0)
    batch = moop4.filterBatch(batch)
    for (x, i) in batch:
        moop4.evaluateSimulation(x, i)
    moop4.updateAll(0, batch)
    batch = moop4.iterate(1)
    batch = moop4.filterBatch(batch)
    for (x, i) in batch:
        moop4.evaluateSimulation(x, i)
    moop4.updateAll(1, batch)
    soln = moop4.getPF()
    # Assert that solutions were found
    assert (soln.size > 0)
    # Assert that the x_vals and f_vals match
    sim = {}
    for i, xi in enumerate(soln):
        sim["sim1"] = ((xi["x0"] - 1.0) * (xi["x0"] - 1.0) +
                       (xi["x1"]) * (xi["x1"]) + float(xi["x2"]))
        assert (f9(soln[i], sim) - soln['f1'][i] < 1.0e-8 and
                f10(soln[i], sim) - soln['f2'][i] < 1.0e-8)
        assert (xi["x2"] == "0")


def test_MOOP_solve():
    """ Test the MOOP class's solver in objectives.py.

    Perform a test of the MOOP solver class by minimizing a 5 variable,
    biobjective convex function s.t. $x in [0, 1]^n$.

    The correctness of the solutions is difficult to assert , but we can
    assert  that the efficient points map onto the Pareto front, as
    expected.

    """

    from parmoo import MOOP
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF
    from parmoo.acquisitions import UniformWeights, RandomConstraint
    from parmoo.optimizers import LocalSurrogate_PS, GlobalSurrogate_BFGS
    import numpy as np
    import pytest

    # Initialize two simulation groups with 1 output each
    s1 = {'m': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm([x[f"x{key}"]
                                                 for key in [1, 2, 3, 4]])],
          'surrogate': GaussRBF,
          'search_budget': 20}
    s2 = {'m': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm([x[f"x{key}"] - 1
                                                 for key in [1, 2, 3, 4]])],
          'surrogate': GaussRBF,
          'search_budget': 20}
    # Create a MOOP with 4 design variables and 2 simulations
    moop1 = MOOP(LocalSurrogate_PS, hyperparams={'opt_budget': 100,
                                                 'np_random_gen': 0})
    for i in range(4):
        moop1.addDesign({'lb': 0.0, 'ub': 1.0})
    moop1.addSimulation(s1, s2)
    # Now add 2 objectives
    def f1(x, sim): return sim["sim1"]
    def f2(x, sim): return sim["sim2"]
    moop1.addObjective({'obj_func': f1},
                       {'obj_func': f2})
    # Add 3 acquisition functions
    for i in range(3):
        moop1.addAcquisition({'acquisition': UniformWeights})
    # Try to solve several invalid problems/budgets to test error handling
    with pytest.raises(ValueError):
        moop1.solve(-1)
    with pytest.raises(TypeError):
        moop1.solve(2.0)
    # Solve the MOOP with 6 iterations
    moop1.solve(6)
    soln = moop1.getPF()
    # Assert that solutions were found
    assert (soln.size > 0)
    # Assert that the x_vals and f_vals match
    for i in range(soln.shape[0]):
        assert (np.linalg.norm(np.array([s1['sim_func'](soln[i]),
                                         s2['sim_func'](soln[i])]).flatten() -
                               np.array([soln['f1'][i], soln['f2'][i]]))
                < 1.e-8)
    # Create new single objective toy problem
    s3 = {'m': 1,
          'sim_func': lambda x: [x["x1"] + x["x2"]],
          'surrogate': GaussRBF,
          'search': LatinHypercube,
          'hyperparams': {'search_budget': 10}}
    s4 = {'m': 1,
          'sim_func': lambda x: [x["x3"] + x["x4"]],
          'surrogate': GaussRBF,
          'search': LatinHypercube,
          'hyperparams': {'search_budget': 20}}
    moop2 = MOOP(LocalSurrogate_PS, hyperparams={'np_random_gen': 0})
    for i in range(4):
        moop2.addDesign({'lb': 0.0, 'ub': 1.0})
    moop2.addSimulation(s3, s4)
    # Now add 1 objective
    def f3(x, sim): return sim["sim1"][0] + sim["sim2"][0]
    moop2.addObjective({'obj_func': f3})
    # Add 3 acquisition functions
    for i in range(3):
        moop2.addAcquisition({'acquisition': RandomConstraint})
    # Solve the MOOP and extract the final database with 6 iterations
    moop2.solve(6)
    soln = moop2.getPF()
    # Assert that solutions were found
    assert (soln.size > 0)
    # Assert that the x_vals and f_vals match
    for i in range(soln.shape[0]):
        assert (np.linalg.norm(np.array(s3['sim_func'](soln[i])) +
                               np.array(s4['sim_func'](soln[i])) -
                               soln['f1'][i]) < 0.00000001)

    # Create a 3 objective toy problem, with no simulations
    moop3 = MOOP(GlobalSurrogate_BFGS, hyperparams={'np_random_gen': 0})
    for i in range(4):
        moop3.addDesign({'lb': 0.0, 'ub': 1.0})

    # Now add the three objectives
    def f4(x, sim):
        return np.linalg.norm([x["x1"] - 1, x["x2"], x["x3"], x["x4"]]) ** 2.0

    def g4(x, sim):
        dx = {"x1": 2 * x["x1"] - 2,
              "x2": 2 * x["x2"],
              "x3": 2 * x["x3"],
              "x4": 2 * x["x4"]}
        return dx, {}

    def f5(x, sim):
        return np.linalg.norm([x["x1"], x["x2"] - 1, x["x3"], x["x4"]]) ** 2.0

    def g5(x, sim):
        dx = {"x1": 2 * x["x1"],
              "x2": 2 * x["x2"] - 2,
              "x3": 2 * x["x3"],
              "x4": 2 * x["x4"]}
        return dx, {}

    def f6(x, sim):
        return np.linalg.norm([x["x1"], x["x2"], x["x3"] - 1, x["x4"]]) ** 2.0

    def g6(x, sim):
        dx = {"x1": 2 * x["x1"],
              "x2": 2 * x["x2"],
              "x3": 2 * x["x3"] - 2,
              "x4": 2 * x["x4"]}
        return dx, {}

    moop3.addObjective({'obj_func': f4, 'obj_grad': g4},
                       {'obj_func': f5, 'obj_grad': g5},
                       {'obj_func': f6, 'obj_grad': g6})
    # Add 3 acquisition functions
    for i in range(3):
        moop3.addAcquisition({'acquisition': UniformWeights})
    # Solve the MOOP and extract the final database with 6 iterations
    moop3.solve(6)
    soln = moop3.getPF()
    # Assert that solutions were found
    assert (soln.size > 0)
    # Assert that the x_vals and f_vals match
    sim = np.zeros(0)
    for i in range(soln.shape[0]):
        assert (np.linalg.norm(np.array([f4(soln[i], sim),
                                         f5(soln[i], sim),
                                         f6(soln[i], sim)]
                                        ).flatten() -
                               np.array([soln[i]["f1"], soln[i]["f2"],
                                         soln[i]["f3"]])) < 0.00000001)

    # Create a 3 objective toy problem, with no simulations and 1 categorical
    moop4 = MOOP(GlobalSurrogate_BFGS, hyperparams={'np_random_gen': 0})
    for i in range(3):
        moop4.addDesign({'lb': 0.0, 'ub': 1.0})
    moop4.addDesign({'des_type': "categorical", 'levels': 3})
    moop4.addObjective({'obj_func': f4, 'obj_grad': g4},
                       {'obj_func': f5, 'obj_grad': g5},
                       {'obj_func': f6, 'obj_grad': g6})
    # Add 3 acquisition functions
    for i in range(3):
        moop4.addAcquisition({'acquisition': UniformWeights})
    # Solve the MOOP and extract the final database with 6 iterations
    moop4.solve(6)
    soln = moop4.getPF()
    # Assert that solutions were found
    assert (soln.size > 0)
    # Assert that the x_vals and f_vals match
    sim = np.zeros(0)
    for i in range(soln.shape[0]):
        assert (np.linalg.norm(np.array([f4(soln[i], sim),
                                         f5(soln[i], sim),
                                         f6(soln[i], sim)]
                                        ).flatten() -
                               np.array([soln[i]["f1"], soln[i]["f2"],
                                         soln[i]["f3"]])) < 0.00000001)


if __name__ == "__main__":
    test_MOOP_iterate()
    test_MOOP_solve()




DTYPE = [("x1", "f8"), ("x2", "f8"), ("x3", "f8"), ("x4", "f8"), ("x5", "f8")]
ONES = {"x1": 0.6, "x2": 0.6, "x3": 0.6, "x4": 0.6, "x5": 0.6}


def test_sims_dtlz_g1():
    """ Test the g1 kernel function.

    Create an instance of g1 and check that its output is correct.

    """

    from parmoo.simulations.dtlz import g1_sim
    import numpy as np

    g1 = g1_sim(DTYPE, num_obj=3, offset=0.6)
    assert (np.abs(g1(ONES)) < 1.0e-4)


def test_sims_dtlz_g2():
    """ Test the g2 kernel function.

    Create an instance of g2 and check that its output is correct.

    """

    from parmoo.simulations.dtlz import g2_sim
    import numpy as np

    g2 = g2_sim(DTYPE, num_obj=3, offset=0.6)
    assert (np.abs(g2(ONES)) < 1.0e-4)


def test_sims_dtlz_g3():
    """ Test the g3 kernel function.

    Create an instance of g3 and check that its output is correct.

    """

    from parmoo.simulations.dtlz import g3_sim
    import numpy as np

    g3 = g3_sim(DTYPE, num_obj=3, offset=0.6)
    assert (np.abs(g3(ONES)) < 1.0e-4)


def test_sims_dtlz_g4():
    """ Test the g4 kernel function.

    Create an instance of g4 and check that its output is correct.

    """

    from parmoo.simulations.dtlz import g4_sim
    import numpy as np

    g4 = g4_sim(DTYPE, num_obj=3, offset=0.6)
    assert (np.abs(g4(ONES) - 1.0) < 1.0e-4)


def test_sims_dtlz_dtlz1():
    """ Test the dtlz1 simulation function.

    Create an instance of dtlz1_sim and check that its output is correct.

    """

    from parmoo.simulations.dtlz import dtlz1_sim

    dtlz1 = dtlz1_sim(DTYPE, num_obj=3, offset=0.6)
    assert (sum(dtlz1(ONES)) - 0.5 < 1.0e-4)


def test_sims_dtlz_dtlz2():
    """ Test the dtlz2 simulation function.

    Create an instance of dtlz2_sim and check that its output is correct.

    """

    from parmoo.simulations.dtlz import dtlz2_sim
    import numpy as np

    dtlz2 = dtlz2_sim(DTYPE, num_obj=3, offset=0.6)
    assert (np.linalg.norm(dtlz2(ONES)) - 1.0 < 1.0e-4)


def test_sims_dtlz_dtlz3():
    """ Test the dtlz3 simulation function.

    Create an instance of dtlz3_sim and check that its output is correct.

    """

    from parmoo.simulations.dtlz import dtlz3_sim
    import numpy as np

    dtlz3 = dtlz3_sim(DTYPE, num_obj=3, offset=0.6)
    assert (np.linalg.norm(dtlz3(ONES)) - 1.0 < 1.0e-4)


def test_sims_dtlz_dtlz4():
    """ Test the dtlz4 simulation function.

    Create an instance of dtlz4_sim and check that its output is correct.

    """

    from parmoo.simulations.dtlz import dtlz4_sim
    import numpy as np

    dtlz4 = dtlz4_sim(DTYPE, num_obj=3, offset=0.6)
    assert (np.linalg.norm(dtlz4(ONES)) - 1.0 < 1.0e-4)


def test_sims_dtlz_dtlz5():
    """ Test the dtlz5 simulation function.

    Create an instance of dtlz5_sim and check that its output is correct.

    """

    from parmoo.simulations.dtlz import dtlz5_sim
    import numpy as np

    dtlz5 = dtlz5_sim(DTYPE, num_obj=3, offset=0.6)
    assert (np.linalg.norm(dtlz5(ONES)) - 1.0 < 1.0e-4)


def test_sims_dtlz_dtlz6():
    """ Test the dtlz6 simulation function.

    Create an instance of dtlz6_sim and check that its output is correct.

    """

    from parmoo.simulations.dtlz import dtlz6_sim
    import numpy as np

    dtlz6 = dtlz6_sim(DTYPE, num_obj=3, offset=0.6)
    assert (np.linalg.norm(dtlz6(ONES)) - 1.0 < 1.0e-4)


def test_sims_dtlz_dtlz7():
    """ Test the dtlz7 simulation function.

    Create an instance of dtlz7_sim and check that its output is correct.

    """

    from parmoo.simulations.dtlz import dtlz7_sim
    import numpy as np

    dtlz7 = dtlz7_sim(DTYPE, num_obj=3, offset=0.6)
    x_in = ONES.copy()
    x_in["x1"] = 0.0
    x_in["x2"] = 0.0
    assert (np.abs(dtlz7(x_in)[2] - 6.0) < 1.0e-4)


def test_sims_dtlz_dtlz8():
    """ Test the dtlz8 simulation function.

    Create an instance of dtlz8_sim and check that its output is correct.

    """

    from parmoo.simulations.dtlz import dtlz8_sim
    import numpy as np

    dtlz8 = dtlz8_sim(DTYPE, num_obj=3, offset=0.6)
    assert (np.all(np.abs(dtlz8(ONES)) < 1.0e-4))


def test_sims_dtlz_dtlz9():
    """ Test the dtlz9 simulation function.

    Create an instance of dtlz9_sim and check that its output is correct.

    """

    from parmoo.simulations.dtlz import dtlz9_sim
    import numpy as np

    dtlz9 = dtlz9_sim(DTYPE, num_obj=3, offset=0.6)
    assert (np.all(np.abs(dtlz9(ONES)) < 1.0e-4))


if __name__ == "__main__":
    test_sims_dtlz_g1()
    test_sims_dtlz_g2()
    test_sims_dtlz_g3()
    test_sims_dtlz_g4()
    test_sims_dtlz_dtlz1()
    test_sims_dtlz_dtlz2()
    test_sims_dtlz_dtlz3()
    test_sims_dtlz_dtlz4()
    test_sims_dtlz_dtlz5()
    test_sims_dtlz_dtlz6()
    test_sims_dtlz_dtlz7()
    test_sims_dtlz_dtlz8()
    test_sims_dtlz_dtlz9()




""" Implementations of the SurrogateFunction class.

This module contains implementations of the SurrogateFunction ABC, which rely
on Gaussian basis functions (i.e., Gaussian processes).

The classes include:
 * ``GaussRBF`` -- fits Gaussian radial basis functions (RBFs)

"""

from jax import jit, vmap
from jax import numpy as jnp
from jax import lax
import numpy as np
from parmoo.structs import SurrogateFunction
from parmoo.util import xerror
from scipy.stats import tstd


class GaussRBF(SurrogateFunction):
    """ A RBF surrogate model, using a Gaussian basis.

    This class implements a local RBF surrogate with a Gaussian basis,
    using the SurrogateFunction ABC.

    """

    # Slots for the UniformRandom class
    __slots__ = ['m', 'n', 'lb', 'ub', 'x_vals', 'f_vals', 'eps', 'nugget',
                 'loc_inds', 'tr_center', 'weights', 'prior', 'v', 'w',
                 'order', 'x_std_dev', 'y_std_dev', 'mu']

    def __init__(self, m, lb, ub, hyperparams):
        """ Constructor for the GaussRBF class.

        Args:
            m (int): The number of objectives to fit.

            lb (numpy.ndarray): A 1d array of lower bounds for the design
                region. The number of design variables is inferred from the
                dimension of lb.

            ub (numpy.ndarray): A 1d array of upper bounds for the design
                region. The dimension must match ub.

            hyperparams (dict): A dictionary of hyperparameters for the
                RBF models, including:
                 * des_tols (numpy.ndarray, optional): A 1d array whose length
                   matches lb and ub. Each entry is a number (greater than 0)
                   specifying the design space tolerance for that variable.
                   By default, des_tols = [1.0e-8, ..., 1.0e-8].
                 * tail_order (int, optional): Order of the polynomial tail.
                   Can be 0 or 1, defaults to 0.

        Returns:
            GaussRBF: A new GaussRBF object.

        """

        # Check inputs
        xerror(o=m, lb=lb, ub=ub, hyperparams=hyperparams)
        # Initialize problem dimensions
        self.m = m
        self.lb = lb
        self.ub = ub
        self.n = self.lb.size
        self.x_std_dev = 0.0
        # Create empty database
        self.x_vals = np.zeros((0, self.n))
        self.f_vals = np.zeros((0, self.m))
        self.weights = np.zeros((0, 0))
        self.prior = np.zeros((self.n+1, self.m))
        self.v = np.zeros((0, 0))
        self.w = np.zeros((0, 0))
        self.y_std_dev = np.ones(self.m)
        # Initialize trust-region settings
        self.tr_center = np.zeros(0)
        self.loc_inds = []
        # Check for the 'nugget' optional value in hyperparams
        if 'nugget' in hyperparams:
            if isinstance(hyperparams['nugget'], float):
                self.nugget = hyperparams['nugget']
                if self.nugget < 0.0:
                    raise ValueError("hyperparams['nugget'] cannot be a"
                                     + " negative number")
            else:
                raise ValueError("hyperparams['nugget'] contained an illegal"
                                 + " value")
        else:
            self.nugget = 0.0
        # Check for 'des_tols' optional key in hyperparams
        self.mu = np.sqrt(jnp.finfo(jnp.ones(1)).eps)
        if 'des_tols' in hyperparams:
            if isinstance(hyperparams['des_tols'], np.ndarray):
                if hyperparams['des_tols'].size == self.n:
                    if np.all(hyperparams['des_tols'] > 0.0):
                        self.eps = hyperparams['des_tols']
                    else:
                        raise ValueError("hyperparams['des_tols'] must all be"
                                         + " greater than 0")
                else:
                    raise ValueError("hyperparams['des_tols'] must have length"
                                     + " n")
            else:
                raise ValueError("hyperparams['des_tols'] contained an illegal"
                                 + " value")
        else:
            self.eps = np.zeros(self.n)
            self.eps[:] = self.mu
        # Check for 'tail_order' optional key in hyperparams
        if 'tail_order' in hyperparams:
            if isinstance(hyperparams['tail_order'], int):
                if hyperparams['tail_order'] in [0, 1]:
                    self.order = hyperparams['tail_order']
                else:
                    raise ValueError("hyperparams['tail_order'] must be "
                                     + "0 or 1")
            else:
                raise ValueError("hyperparams['tail_order'] contained an "
                                 + "illegal value")
        else:
            self.order = 0
        return

    def fit(self, x, f):
        """ Fit a new Gaussian RBF to the given data.

        Args:
             x (numpy.ndarray): A 2d array containing the list of
                 design points.

             f (numpy.ndarray): A 2d array containing the corresponding list
                 of objective values.

        """

        # Check that the x and f values are legal
        if not isinstance(x, np.ndarray):
            raise TypeError("x must be a numpy array")
        if isinstance(f, np.ndarray):
            if f.shape[0] == 0:
                raise ValueError("no data provided")
            if self.m != np.size(f[0, :]):
                raise ValueError("each row of f must have length m")
        else:
            raise TypeError("f must be a numpy array")
        if x.shape[0] != f.shape[0]:
            raise ValueError("x and f must have equal lengths")
        # Initialize the internal database with x and f
        self.x_vals = x
        self.f_vals = f
        # Initialize the local indices for future usage
        self.tr_center = self.lb.copy() - 1
        return

    def update(self, x, f):
        """ Update an existing Gaussian RBF using new data.

        Args:
             x (numpy.ndarray): A 2d array containing the list of
                 new design points, with which to update the surrogate
                 models.

             f (numpy.ndarray): A 2d array containing the corresponding list
                 of objective values.

        """

        # Check that the x and f values are legal
        if not isinstance(x, np.ndarray):
            raise TypeError("x must be a numpy array")
        if isinstance(f, np.ndarray):
            if f.shape[0] == 0:
                return    # No new data, do nothing
            if self.m != np.size(f[0, :]):
                raise ValueError("each row of f must have length m")
        else:
            raise TypeError("f must be a numpy array")
        if x.shape[0] != f.shape[0]:
            raise ValueError("x and f must have equal lengths")
        # Update the internal database with x and f
        self.x_vals = np.concatenate((self.x_vals, x), axis=0)
        self.f_vals = np.concatenate((self.f_vals, f), axis=0)
        # Reinitialize the local indices for future usage
        self.tr_center = self.lb.copy() - 1
        return

    def setTrustRegion(self, center, radius):
        """ Set the new trust-region center and refit the local RBF.

        Args:
            center (numpy.ndarray): A 1d array containing the new trust-
                region center.

            radius (numpy.ndarray or float): The trust-region radius.

        """

        # Check that the center is legal
        if not isinstance(center, np.ndarray):
            raise TypeError("center must be a numpy array")
        else:
            if center.size != self.n:
                raise ValueError("center must have length n")
            elif (np.any(center < self.lb - self.eps) or
                  np.any(center > self.ub + self.eps)):
                raise ValueError("center cannot be infeasible")
        # Check that the radius is legal
        if isinstance(radius, np.ndarray):
            if radius.size != self.n:
                raise ValueError("radius must have length n")
            elif np.any(radius <= 0):
                raise ValueError("radius must be positive")
        elif isinstance(radius, float):
            if radius <= 0:
                raise ValueError("radius must be positive")
        else:
            raise TypeError("radius must be a numpy array or float")
        # Some helper variables used below
        rad_tmp = np.array(radius)
        refit = False
        # If the radius is infinite, fit with all data
        if np.all(rad_tmp == np.inf):
            # Only need to refit once after an update
            if np.all(self.tr_center < self.lb):
                # Set the trust-region center and radius to large values
                self.tr_center = (self.ub + self.lb) / 2.0
                tr_radius = (self.ub - self.lb) / 2.0
                # Compute the standard deviation for the Gaussian bubbles
                self.x_std_dev = np.power(np.prod(tr_radius * 2.0) /
                                          np.float64(self.x_vals.shape[0]),
                                          1.0 / np.float64(self.n))
                # Use all points in the current database
                self.loc_inds = [i for i in range(self.x_vals.shape[0])]
                refit = True
        # Otherwise, if the nearest neighbor has changed, refit the RBF
        elif np.any(self.tr_center != center):
            # Update the trust-region center and radius
            self.tr_center = center
            tr_radius = rad_tmp
            # Update the standard deviation for the Gaussian bubbles
            self.x_std_dev = np.linalg.norm(tr_radius * 2)
            # Get points in the new trust region
            self.loc_inds = []
            rdists = np.asarray([np.linalg.norm((xj - center))
                                 for xj in self.x_vals])
            for i, ri in enumerate(rdists):
                if ri < 3.0 * self.x_std_dev:
                    self.loc_inds.append(i)
            refit = True
        # Only do the following if we are re-fitting the models
        if refit:
            cov = _gaussian(_pdist(self.x_vals[self.loc_inds]),
                            self.x_std_dev)
            # Add the nugget, if present
            if self.nugget > 0:
                for i in range(len(self.loc_inds)):
                    cov = cov.at[i, i].set(cov[i, i] + self.nugget)
            # Get eigenvalue decomp to solve the SPD system with multiple RHS
            self.w = np.zeros(cov.shape[0])
            self.v = np.zeros(cov.shape)
            self.w, self.v = np.linalg.eigh(cov)
            # Check the smallest singular value for a bad solution
            sigma_n = np.min(self.w)
            if sigma_n < self.mu:
                for i in range(len(self.loc_inds)):
                    cov = cov.at[i, i].set(cov[i, i] + self.mu - sigma_n)
                self.w, self.v = np.linalg.eigh(cov)
            # Fit prior weights and remove tail effects from RHS
            rhs = self.f_vals[self.loc_inds, :].copy()
            if self.order >= 0:
                self.prior[-1, :] = np.sum(rhs, axis=0) / rhs.shape[0]
                rhs[:, :] = rhs[:, :] - self.prior[-1, :]
                if self.order >= 1:
                    A = self.x_vals[self.loc_inds, :].copy()
                    b = rhs.copy()
                    self.prior[:-1, :] = np.linalg.lstsq(A, b, rcond=None)[0]
                    rhs[:, :] = rhs[:, :] - np.dot(self.x_vals[self.loc_inds],
                                                   self.prior[:-1])
            if rhs.shape[0] > 1:
                self.y_std_dev = tstd(rhs, axis=0)
            else:
                self.y_std_dev = np.ones(rhs.shape[1])
            # Finish the solve
            self.weights = np.zeros((self.m, rhs.shape[0]))
            for i in range(self.m):
                tmp = np.dot(self.v.T, rhs[:, i]) / self.w[:]
                self.weights[i, :] = np.dot(self.v, tmp)
        return

    def evaluate(self, x):
        """ Evaluate the Gaussian RBF at a design point.

        Args:
            x (numpy.ndarray): A 1d array containing the design point at
                which to the Gaussian RBF should be evaluated.

        Returns:
            numpy.ndarray: A 1d array containing the predicted objective value
            at x.

        """

        return lax.cond(float(self.order) < 1, _evaluate_0, _evaluate_1,
                        self.x_vals[self.loc_inds], self.x_std_dev,
                        self.weights, self.prior, x)

    def stdDev(self, x):
        """ Evaluate the standard deviation of the Gaussian RBF at x.

        Args:
            x (numpy.ndarray): A 1d array containing the design point at
                which the standard deviation should be evaluated.

        Returns:
            numpy.ndarray: A 1d array containing the standard deviation at x.

        """

        return _evaluate_sd(self.x_vals[self.loc_inds], self.v, self.w,
                            self.x_std_dev, self.mu, x) * self.y_std_dev

    def save(self, filename):
        """ Save important data from this class so that it can be reloaded.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data should be saved.

        """

        import json

        # Serialize RBF object in dictionary
        gp_state = {'m': self.m,
                    'n': self.n,
                    'x_std_dev': self.x_std_dev,
                    'nugget': self.nugget,
                    'loc_inds': self.loc_inds,
                    'order': self.order}
        # Serialize numpy.ndarray objects
        gp_state['lb'] = self.lb.tolist()
        gp_state['ub'] = self.ub.tolist()
        gp_state['x_vals'] = self.x_vals.tolist()
        gp_state['f_vals'] = self.f_vals.tolist()
        gp_state['eps'] = self.eps.tolist()
        gp_state['tr_center'] = self.tr_center.tolist()
        gp_state['weights'] = self.weights.tolist()
        gp_state['prior'] = self.prior.tolist()
        gp_state['v'] = self.v.tolist()
        gp_state['w'] = self.w.tolist()
        gp_state['y_std_dev'] = self.y_std_dev.tolist()
        # Save file
        with open(filename, 'w') as fp:
            json.dump(gp_state, fp)
        return

    def load(self, filename):
        """ Reload important data into this class after a previous save.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data has been saved.

        """

        import json

        # Load file
        with open(filename, 'r') as fp:
            gp_state = json.load(fp)
        # Deserialize RBF object from dictionary
        self.m = gp_state['m']
        self.n = gp_state['n']
        self.x_std_dev = gp_state['x_std_dev']
        self.nugget = gp_state['nugget']
        self.loc_inds = gp_state['loc_inds']
        self.order = gp_state['order']
        # Deserialize numpy.ndarray objects
        self.lb = np.array(gp_state['lb'])
        self.ub = np.array(gp_state['ub'])
        self.x_vals = np.array(gp_state['x_vals'])
        self.f_vals = np.array(gp_state['f_vals'])
        self.eps = np.array(gp_state['eps'])
        self.tr_center = np.array(gp_state['tr_center'])
        self.weights = np.array(gp_state['weights'])
        self.prior = np.array(gp_state['prior'])
        self.v = np.array(gp_state['v'])
        self.w = np.array(gp_state['w'])
        self.y_std_dev = np.array(gp_state['y_std_dev'])
        return


# Private pure helper functions


@jit
def _gaussian(r2, x_std_dev):
    """ Evaluate Gaussian bump with x_std_dev at distance r^2 from center """

    return jnp.exp(-r2 / (x_std_dev ** 2))


@jit
def _cdist(x_vals, x):
    """ Compute all squared distances from points in x_vals to x """

    return jnp.sum((x_vals - x) ** 2, axis=1)


@jit
def _pdist(x_vals):
    """ Compute all pairwise squared distances to the input arg """

    return vmap(lambda x: _cdist(x_vals, x))(x_vals)


@jit
def _evaluate_0(x_vals, x_std_dev, weights, prior, x):
    """ Evaluate a Gaussian RBF (constant prior) at a design point x. """

    post_tmp = jnp.dot(weights, _gaussian(_cdist(x_vals, x), x_std_dev))
    return post_tmp + prior[-1, :]


@jit
def _evaluate_1(x_vals, x_std_dev, weights, prior, x):
    """ Evaluate a Gaussian RBF (linear prior) at a design point x. """

    post_tmp = jnp.dot(weights, _gaussian(_cdist(x_vals, x), x_std_dev))
    pre_tmp = jnp.dot(x, prior[:-1, :]) + prior[-1, :]
    return post_tmp + pre_tmp


@jit
def _evaluate_sd(x_vals, v, w, x_std_dev, eps, x):
    """ Evaluate the posterior standard deviation of a Gaussian RBF at x. """

    vTc = jnp.dot(v.T, _gaussian(_cdist(x_vals, x), x_std_dev))
    return jnp.sqrt(jnp.maximum(1 - jnp.dot(vTc / w, vTc.T), eps))




""" Optimization methods based on limited-memory BFGS-B (L-BFGS-B).

This module contains implementations of the SurrogateOptimizer ABC, which
are based on the L-BFGS-B quasi-Newton algorithm.

Note that all of these methods are gradient based, and therefore require
objective, constraint, and surrogate gradient methods to be defined.

The classes include:
 * ``GlobalSurrogate_BFGS`` -- Minimize the surrogate globally via L-BFGS-B
 * ``LocalSurrogate_BFGS`` -- Minimize surrogate in trust region via L-BFGS-B

"""

from jax import config
import jax
from jax import numpy as jnp
import numpy as np
from parmoo.structs import SurrogateOptimizer
from parmoo.util import xerror
config.update("jax_enable_x64", True)  # scipy.optimize.lbfgsb requires 64-bit


class GlobalSurrogate_BFGS(SurrogateOptimizer):
    """ Use L-BFGS-B and gradients to identify local solutions.

    Applies L-BFGS-B to the surrogate problem, in order to identify design
    points that are locally Pareto optimal with respect to the surrogate
    problem.

    """

    # Slots for the GlobalSurrogate_BFGS class
    __slots__ = ['n', 'bounds', 'acquisitions', 'budget', 'constraints',
                 'objectives', 'simulations', 'setTR', 'penalty_func',
                 'sim_sd', 'np_rng', 'mu']

    def __init__(self, o, lb, ub, hyperparams):
        """ Constructor for the GlobalSurrogate_BFGS class.

        Args:
            o (int): The number of objectives.

            lb (numpy.ndarray): A 1d array of lower bounds for the design
                region. The number of design variables is inferred from the
                dimension of lb.

            ub (numpy.ndarray): A 1d array of upper bounds for the design
                region. The dimension must match ub.

            hyperparams (dict): A dictionary of hyperparameters for the
                optimization procedure. It may contain the following:
                 * opt_budget (int): The iteration limit per solve
                   (default: 100).
                 * opt_restarts (int): Number of multistart restarts per
                   scalarization (default: n+1).

        Returns:
            SurrogateOptimizer: A new SurrogateOptimizer object.

        """

        # Check inputs
        xerror(o=o, lb=lb, ub=ub, hyperparams=hyperparams)
        self.n = lb.size
        self.bounds = np.zeros((self.n, 2))
        self.bounds[:, 0] = lb
        self.bounds[:, 1] = ub
        self.mu = np.sqrt(jnp.finfo(jnp.ones(1)).eps)
        # Check that the contents of hyperparams is legal
        if 'opt_restarts' in hyperparams:
            if isinstance(hyperparams['opt_restarts'], int):
                if hyperparams['opt_restarts'] < 1:
                    raise ValueError("hyperparams['opt_restarts'] "
                                     "must be positive")
                else:
                    self.restarts = hyperparams['opt_restarts']
            else:
                raise TypeError("hyperparams['opt_restarts'] "
                                "must be an integer")
        else:
            self.restarts = self.n + 1
        if 'opt_budget' in hyperparams:
            if isinstance(hyperparams['opt_budget'], int):
                if hyperparams['opt_budget'] < 1:
                    raise ValueError("hyperparams['opt_budget'] "
                                     "must be positive")
                else:
                    self.budget = hyperparams['opt_budget']
            else:
                raise TypeError("hyperparams['opt_budget'] "
                                "must be an integer")
        else:
            self.budget = 100
        # Check the hyperparameter dictionary for random generator
        if 'np_random_gen' in hyperparams:
            if isinstance(hyperparams['np_random_gen'], np.random.Generator):
                self.np_rng = hyperparams['np_random_gen']
            else:
                raise TypeError("When present, hyperparams['np_random_gen'] "
                                "must be an instance of the class "
                                "numpy.random.Generator")
        else:
            self.np_rng = np.random.default_rng()
        self.acquisitions = []
        return

    def solve(self, x):
        """ Solve the surrogate problem using L-BFGS-B.

        Args:
            x (np.ndarray): A 2d array containing a list of design points
                used to warm start the search.

        Returns:
            np.ndarray: A 2d numpy.ndarray of potentially efficient design
            points that were found by L-BFGS-B.

        """

        from scipy import optimize

        # Check that x is legal
        if self.n != x.shape[1]:
            raise ValueError("The columns of x must match n")
        elif len(self.acquisitions) != x.shape[0]:
            raise ValueError("The rows of x must match the number " +
                             "of acquisition functions")
        # Check that x is feasible.
        for xj in x:
            if np.any(xj[:] < self.bounds[:, 0]) or \
               np.any(xj[:] > self.bounds[:, 1]):
                raise ValueError("some of starting points (x) are infeasible")
        # Create an infinite trust region
        rad = np.ones(self.n) * np.inf
        self.setTR(np.zeros(self.n), rad)
        # Loop over and solve acquisition functions
        result = []
        for j, acquisition in enumerate(self.acquisitions):

            # Define the scalarized wrapper functions
            if acquisition.useSD():

                def _scal_f(x, *args):
                    sx = self.simulations(x)
                    sdx = self.sim_sd(x)
                    fx = self.penalty_func(x, sx)
                    return acquisition.scalarize(fx, x, sx, sdx)

            else:

                def _scal_f(x, *args):
                    sx = self.simulations(x)
                    sdx = jnp.zeros(sx.size)
                    fx = self.penalty_func(x, sx)
                    return acquisition.scalarize(fx, x, sx, sdx)

            # Recompile the scalar functions
            try:
                scalar_f = jax.jit(_scal_f)
                x0 = x[j, :].copy()
                _ = scalar_f(x0)
            except BaseException:
                scalar_f = _scal_f
                x0 = x[j, :].copy()
                _ = scalar_f(x0)
            try:
                _scal_g = jax.jit(jax.jacrev(_scal_f))
                _ = _scal_g(x0)
            except BaseException:
                _scal_g = jax.jacrev(_scal_f)

            def scalar_g(x, *ag): return np.asarray(_scal_g(x, *ag)).flatten()

            g0 = scalar_g(x0)
            # Get the solution via multistart solve
            soln = x[j, :].copy()
            for i in range(self.restarts):
                if i == 1:
                    # Use predicted gradient step to warm-start second start
                    for ii in range(self.n):
                        if g0[ii] < -1.0e-8:
                            x0[ii] = self.bounds[ii, 1]
                        elif g0[ii] > 1.0e-8:
                            x0[ii] = self.bounds[ii, 0]
                elif i > 1:
                    # Random starting point within bounds for all other starts
                    x0 = (self.np_rng.random(self.n) *
                          (self.bounds[:, 1] - self.bounds[:, 0]) +
                          self.bounds[:, 0])
                # Solve the problem globally within bound constraints
                res = optimize.minimize(scalar_f, x0, method='L-BFGS-B',
                                        jac=scalar_g, bounds=self.bounds,
                                        options={'maxiter': self.budget,
                                                 'ftol': self.mu,
                                                 'gtol': np.sqrt(self.mu)})
                if scalar_f(res['x']) < scalar_f(soln):
                    soln = res['x']
            # Append the found minima to the results list
            result.append(soln)
        self.objectives = None
        self.constraints = None
        self.penalty_func = None
        return np.asarray(result)


class LocalSurrogate_BFGS(SurrogateOptimizer):
    """ Use L-BFGS-B and gradients to identify solutions within a trust region.

    Applies L-BFGS-B to the surrogate problem, in order to identify design
    points that are locally Pareto optimal with respect to the surrogate
    problem.

    """

    # Slots for the LBFGSB class
    __slots__ = ['n', 'bounds', 'acquisitions', 'budget', 'constraints',
                 'objectives', 'penalty_func', 'setTR',
                 'restarts', 'simulations', 'sim_sd', 'prev_centers',
                 'des_tols', 'targets', 'np_rng', 'mu']

    def __init__(self, o, lb, ub, hyperparams):
        """ Constructor for the LocalSurrogate_BFGS class.

        Args:
            o (int): The number of objectives.

            lb (numpy.ndarray): A 1d array of lower bounds for the design
                region. The number of design variables is inferred from the
                dimension of lb.

            ub (numpy.ndarray): A 1d array of upper bounds for the design
                region. The dimension must match ub.

            hyperparams (dict): A dictionary of hyperparameters for the
                optimization procedure. It may contain the following:
                 * opt_budget (int): The iteration limit per solve
                   (default: 500).
                 * opt_restarts (int): Number of multistart restarts per
                   scalarization (default: 2).

        Returns:
            SurrogateOptimizer: A new SurrogateOptimizer object.

        """

        # Check inputs
        xerror(o=o, lb=lb, ub=ub, hyperparams=hyperparams)
        self.n = lb.size
        self.bounds = np.zeros((self.n, 2))
        self.bounds[:, 0] = lb
        self.bounds[:, 1] = ub
        # Check that the contents of hyperparams is legal
        if 'opt_budget' in hyperparams:
            if isinstance(hyperparams['opt_budget'], int):
                if hyperparams['opt_budget'] < 1:
                    raise ValueError("hyperparams['opt_budget'] "
                                     "must be positive")
                else:
                    self.budget = hyperparams['opt_budget']
            else:
                raise TypeError("hyperparams['opt_budget'] "
                                "must be an integer")
        else:
            self.budget = 500
        # Check that the contents of hyperparams is legal
        if 'opt_restarts' in hyperparams:
            if isinstance(hyperparams['opt_restarts'], int):
                if hyperparams['opt_restarts'] < 1:
                    raise ValueError("hyperparams['opt_restarts'] "
                                     "must be positive")
                else:
                    self.restarts = hyperparams['opt_restarts']
            else:
                raise TypeError("hyperparams['opt_restarts'] "
                                "must be an integer")
        else:
            self.restarts = 2
        self.mu = np.sqrt(jnp.finfo(jnp.ones(1)).eps)
        if 'des_tols' in hyperparams:
            if isinstance(hyperparams['des_tols'], np.ndarray):
                if hyperparams['des_tols'].size != self.n:
                    raise ValueError("the length of hyperparpams['des_tols']"
                                     " must match the length of lb and ub")
                if not np.all(hyperparams['des_tols']):
                    raise ValueError("all entries in hyperparams['des_tols']"
                                     " must be greater than 0")
            else:
                raise TypeError("hyperparams['des_tols'] must be an array.")
            self.des_tols = np.asarray(hyperparams['des_tols'])
        else:
            self.des_tols = (np.ones(self.n) * self.mu)
        # Check the hyperparameter dictionary for random generator
        if 'np_random_gen' in hyperparams:
            if isinstance(hyperparams['np_random_gen'], np.random.Generator):
                self.np_rng = hyperparams['np_random_gen']
            else:
                raise TypeError("When present, hyperparams['np_random_gen'] "
                                "must be an instance of the class "
                                "numpy.random.Generator")
        else:
            self.np_rng = np.random.default_rng()
        self.acquisitions = []
        self.prev_centers = []
        self.targets = []
        return

    def __checkTR(self, center):
        """ Check the recommended trust region for a new center. """

        # Search the history for the given radius
        rad = np.zeros(self.n)
        for (ci, ri) in reversed(self.prev_centers):
            if np.all(np.abs(center - np.asarray(ci)) < self.des_tols):
                rad[:] = np.asarray(ri)
                break
        # If not found in the history initialize
        if np.all(rad == 0):
            rad = (self.bounds[:, 1] - self.bounds[:, 0]) * 0.1
            rad = np.maximum(rad, self.des_tols)
        return rad

    def __checkTargets(self):
        """ Use internal list of targets to check and update the TR radii """

        for ti in self.targets:
            # Decay all "missed" targets' TR radii
            ci, ri, _, _ = ti
            ri = np.maximum(ri * 0.5, self.des_tols)
            # Update the TR history
            found = False
            j = 0
            for (cj, rj) in reversed(self.prev_centers):
                j -= 1
                if np.all(np.abs(cj - ci) < self.des_tols):
                    self.prev_centers[j][-1] = ri
            if not found:
                self.prev_centers.append([ci, ri])
        # Reset the list of targets for next iteration
        self.targets = []
        return

    def returnResults(self, x, fx, sx, sdx):
        """ Collect the results of a function evaluation.

        Args:
            x (np.ndarray): The design point evaluated.

            fx (np.ndarray): The objective function values at x.

            sx (np.ndarray): The simulation function values at x.

            sdx (np.ndarray): The standard deviation in the simulation
                outputs at x.

        """

        for i, ti in enumerate(self.targets):
            j = ti[3]
            fxj = self.acquisitions[j].scalarize(fx, x, sx, sdx)
            # Remove any targets that have been "hit"
            if fxj < ti[2]:
                del self.targets[i]
        return

    def solve(self, x):
        """ Solve the surrogate problem using L-BFGS-B.

        Args:
            x (np.ndarray): A 2d array containing a list of design points
                used to warm start the search.

        Returns:
            np.ndarray: A 2d numpy.ndarray of potentially efficient design
            points that were found by L-BFGS-B.

        """

        from scipy import optimize

        # Check that x is legal
        if self.n != x.shape[1]:
            raise ValueError("The columns of x must match n")
        elif len(self.acquisitions) != x.shape[0]:
            raise ValueError("The rows of x must match the number " +
                             "of acquisition functions")
        # Check that x is feasible.
        for xj in x:
            if np.any(xj[:] < self.bounds[:, 0]) or \
               np.any(xj[:] > self.bounds[:, 1]):
                raise ValueError("some of starting points (x) are infeasible")
        # Reset targets and decay any trust regions from previous iteration
        self.__checkTargets()
        # Initialize an empty list of results
        result = []
        # For each acquisition function
        for j, acquisition in enumerate(self.acquisitions):

            # Define the scalarized wrapper functions
            if acquisition.useSD():

                def _scal_f(x, *args):
                    sx = self.simulations(x)
                    sdx = self.sim_sd(x)
                    fx = self.penalty_func(x, sx)
                    return acquisition.scalarize(fx, x, sx, sdx)

            else:

                def _scal_f(x, *args):
                    sx = self.simulations(x)
                    sdx = jnp.zeros(sx.size)
                    fx = self.penalty_func(x, sx)
                    return acquisition.scalarize(fx, x, sx, sdx)

            # Create a new trust region
            rad = self.__checkTR(x[j, :])
            self.setTR(x[j, :], rad)
            bounds = np.zeros((self.n, 2))
            bounds[:, 0] = np.maximum(self.bounds[:, 0], x[j, :] - rad)
            bounds[:, 1] = np.minimum(self.bounds[:, 1], x[j, :] + rad)
            # Recompile the scalar functions
            try:
                scalar_f = jax.jit(_scal_f)
                x0 = x[j, :].copy()
                _ = scalar_f(x0)
            except BaseException:
                scalar_f = _scal_f
                x0 = x[j, :].copy()
                _ = scalar_f(x0)
            try:
                _scal_g = jax.jit(jax.jacrev(_scal_f))
                _ = _scal_g(x0)
            except BaseException:
                _scal_g = jax.jacrev(_scal_f)

            def scalar_g(x, *ag): return np.asarray(_scal_g(x, *ag)).flatten()

            g0 = scalar_g(x0)
            # Get the solution via multistart solve
            soln = x[j, :].copy()
            _ = scalar_f(soln)
            for i in range(self.restarts):
                if i == 1:
                    # Use predicted gradient step to warm-start second start
                    for ii in range(self.n):
                        if g0[ii] < -1.0e-8:
                            x0[ii] = bounds[ii, 1]
                        elif g0[ii] > 1.0e-8:
                            x0[ii] = bounds[ii, 0]
                elif i > 1:
                    # Random starting point within bounds for all other starts
                    x0 = (self.np_rng.random(self.n) *
                          (bounds[:, 1] - bounds[:, 0]) + bounds[:, 0])
                # Solve the problem within the local trust region
                res = optimize.minimize(scalar_f, x0, method='L-BFGS-B',
                                        jac=scalar_g, bounds=bounds,
                                        options={'maxiter': self.budget,
                                                 'ftol': self.mu,
                                                 'gtol': np.sqrt(self.mu)})
                xj = res['x']
                fj = scalar_f(res['x'])
                if fj < scalar_f(soln):
                    soln = xj
            # Append the found minima to the results list
            result.append(soln)
            # We need to remember this "target" for later
            self.targets.append([x[j, :], rad, float(fj), j])
        self.objectives = None
        self.constraints = None
        self.penalty_func = None
        return np.asarray(result)

    def save(self, filename):
        """ Save important data from this class so that it can be reloaded.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data should be saved.

        """

        import json

        # Serialize BFGS object in dictionary
        bfgs_state = {'n': self.n,
                      'budget': self.budget,
                      'restarts': self.restarts}
        # Serialize numpy.ndarray objects
        bfgs_state['bounds'] = self.bounds.tolist()
        bfgs_state['des_tols'] = self.des_tols.tolist()
        # Flatten arrays
        bfgs_state['prev_centers'] = []
        for (ci, ri) in self.prev_centers:
            bfgs_state['prev_centers'].append([ci.tolist(), ri.tolist()])
        bfgs_state['targets'] = []
        for ti in self.targets:
            bfgs_state['targets'].append([ti[0].tolist(), ti[1].tolist(),
                                          ti[2], ti[3]])
        # Save file
        with open(filename, 'w') as fp:
            json.dump(bfgs_state, fp)
        return

    def load(self, filename):
        """ Reload important data into this class after a previous save.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data has been saved.

        """

        import json

        # Load file
        with open(filename, 'r') as fp:
            bfgs_state = json.load(fp)
        # Deserialize BFGS object from dictionary
        self.n = bfgs_state['n']
        self.budget = bfgs_state['budget']
        self.restarts = bfgs_state['restarts']
        # Deserialize numpy.ndarray objects
        self.bounds = np.array(bfgs_state['bounds'])
        self.des_tols = np.array(bfgs_state['des_tols'])
        # Extract history arrays
        self.prev_centers = []
        for (ci, ri) in bfgs_state['prev_centers']:
            self.prev_centers.append([np.array(ci), np.array(ri)])
        self.targets = []
        for ti in bfgs_state['targets']:
            self.targets.append([np.array(ti[0]),
                                 np.array(ti[1]), ti[2], ti[3]])
        return




""" Implementations of the SurrogateFunction class.

This module contains implementations of the SurrogateFunction ABC, which rely
on local linear models.

The classes include:
 * ``Linear`` -- fits a local linear surrogate

"""

from jax import numpy as jnp
import numpy as np
from parmoo.structs import SurrogateFunction
from parmoo.util import xerror


class Linear(SurrogateFunction):
    """ A local linear surrogate model.

    This class implements a local (TR constrained) linear surrogate,
    using the SurrogateFunction ABC.

    """

    # Slots for the UniformRandom class
    __slots__ = ['m', 'n', 'lb', 'ub', 'x_vals', 'f_vals', 'eps',
                 'loc_inds', 'tr_center', 'weights']

    def __init__(self, m, lb, ub, hyperparams):
        """ Constructor for the Linear class.

        Args:
            m (int): The number of objectives to fit.

            lb (numpy.ndarray): A 1d array of lower bounds for the design
                region. The number of design variables is inferred from the
                dimension of lb.

            ub (numpy.ndarray): A 1d array of upper bounds for the design
                region. The dimension must match ub.

            hyperparams (dict): A dictionary of hyperparameters for the
                RBF models, including:
                 * des_tols (numpy.ndarray, optional): A 1d array whose length
                   matches lb and ub. Each entry is a number (greater than 0)
                   specifying the design space tolerance for that variable.
                   By default, des_tols = [1.0e-8, ..., 1.0e-8].

        Returns:
            Linear: A new Linear surrogate object.

        """

        # Check inputs
        xerror(o=m, lb=lb, ub=ub, hyperparams=hyperparams)
        # Initialize problem dimensions
        self.m = m
        self.lb = lb
        self.ub = ub
        self.n = self.lb.size
        # Create empty database
        self.x_vals = np.zeros((0, self.n))
        self.f_vals = np.zeros((0, self.m))
        self.weights = np.zeros(self.n + 1)
        # Initialize trust-region settings
        self.tr_center = np.zeros(0)
        self.loc_inds = []
        # Check for 'des_tols' optional key in hyperparams
        mu = np.sqrt(jnp.finfo(jnp.ones(1)).eps)
        if 'des_tols' in hyperparams:
            if isinstance(hyperparams['des_tols'], np.ndarray):
                if hyperparams['des_tols'].size == self.n:
                    if np.all(hyperparams['des_tols'] > 0.0):
                        self.eps = hyperparams['des_tols']
                    else:
                        raise ValueError("hyperparams['des_tols'] must all be"
                                         + " greater than 0")
                else:
                    raise ValueError("hyperparams['des_tols'] must have length"
                                     + " n")
            else:
                raise ValueError("hyperparams['des_tols'] contained an illegal"
                                 + " value")
        else:
            self.eps = np.zeros(self.n)
            self.eps[:] = mu
        return

    def fit(self, x, f):
        """ Fit a new linear model to the given data.

        Args:
             x (numpy.ndarray): A 2d array containing the list of
                 design points.

             f (numpy.ndarray): A 2d array containing the corresponding list
                 of objective values.

        """

        # Check that the x and f values are legal
        if not isinstance(x, np.ndarray):
            raise TypeError("x must be a numpy array")
        if isinstance(f, np.ndarray):
            if f.shape[0] == 0:
                raise ValueError("no data provided")
            if self.m != np.size(f[0, :]):
                raise ValueError("each row of f must have length m")
        else:
            raise TypeError("f must be a numpy array")
        if x.shape[0] != f.shape[0]:
            raise ValueError("x and f must have equal lengths")
        # Initialize the internal database with x and f
        self.x_vals = x
        self.f_vals = f
        # Reset the center to trigger a re-fit of the model
        self.tr_center = self.lb[:] - np.ones(self.n)
        return

    def update(self, x, f):
        """ Update an existing linear model using new data.

        Args:
             x (numpy.ndarray): A 2d array containing the list of
                 new design points, with which to update the surrogate
                 models.

             f (numpy.ndarray): A 2d array containing the corresponding list
                 of objective values.

        """

        # Check that the x and f values are legal
        if not isinstance(x, np.ndarray):
            raise TypeError("x must be a numpy array")
        if isinstance(f, np.ndarray):
            if f.shape[0] == 0:
                return    # No new data, do nothing
            if self.m != np.size(f[0, :]):
                raise ValueError("each row of f must have length m")
        else:
            raise TypeError("f must be a numpy array")
        if x.shape[0] != f.shape[0]:
            raise ValueError("x and f must have equal lengths")
        # Update the internal database with x and f
        self.x_vals = np.concatenate((self.x_vals, x), axis=0)
        self.f_vals = np.concatenate((self.f_vals, f), axis=0)
        # Reset the center to trigger a re-fit of the model
        self.tr_center = self.lb - np.ones(self.n)
        return

    def setTrustRegion(self, center, radius):
        """ Set the new trust-region center and refit the local linear model.

        Args:
            center (numpy.ndarray): A 1d array containing the new trust-region
                center.

            radius (numpy.ndarray or float): The trust-region radius.

        """

        # Check that the center is legal
        if not isinstance(center, np.ndarray):
            raise TypeError("center must be a numpy array")
        else:
            if center.size != self.n:
                raise ValueError("center must have length n")
            elif (np.any(center < self.lb - self.eps) or
                  np.any(center > self.ub + self.eps)):
                raise ValueError("center cannot be infeasible")
        # Check that the radius is legal
        if isinstance(radius, np.ndarray):
            if radius.size != self.n:
                raise ValueError("radius must have length n")
            elif np.any(radius <= 0):
                raise ValueError("radius must be positive")
        elif isinstance(radius, float):
            if radius <= 0:
                raise ValueError("radius must be positive")
        else:
            raise TypeError("radius must be a numpy array or float")
        # If the center has changed, refit the model
        if np.any(np.abs(self.tr_center - center) > self.eps):
            # Update the center and sort the nearest neighbors
            self.tr_center = center
            idists = np.argsort(np.asarray([np.linalg.norm(xj - center)
                                            for xj in self.x_vals]))
            # Check the n nearest neighbors
            xn = self.x_vals[idists[self.n]]
            r_tmp = np.linalg.norm(center - xn)
            # Get all points within the radius
            self.loc_inds = [int(i) for i in idists
                             if np.linalg.norm(np.maximum(np.abs(center -
                                               self.x_vals[i]) - self.eps, 0))
                             <= r_tmp]
            # Get (min norm) LS fit
            A = np.hstack((self.x_vals[self.loc_inds],
                           np.ones((len(self.loc_inds), 1))))
            self.weights = np.linalg.lstsq(A, self.f_vals[self.loc_inds],
                                           rcond=None)[0]
        return

    def evaluate(self, x):
        """ Evaluate the linear model at a design point.

        Args:
            x (numpy.ndarray): A 1d array containing the design point at
                which the linear model should be evaluated.

        Returns:
            numpy.ndarray: A 1d array containing the predicted objective value
            at x.

        """

        return jnp.dot(self.weights[:-1].T, x).flatten() + self.weights[-1]

    def save(self, filename):
        """ Save important data from this class so that it can be reloaded.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data should be saved.

        """

        import json

        # Serialize RBF object in dictionary
        ls_state = {'m': self.m,
                    'n': self.n,
                    'loc_inds': self.loc_inds}
        # Serialize numpy.ndarray objects
        ls_state['lb'] = self.lb.tolist()
        ls_state['ub'] = self.ub.tolist()
        ls_state['x_vals'] = self.x_vals.tolist()
        ls_state['f_vals'] = self.f_vals.tolist()
        ls_state['eps'] = self.eps.tolist()
        ls_state['tr_center'] = self.tr_center.tolist()
        ls_state['weights'] = self.weights.tolist()
        # Save file
        with open(filename, 'w') as fp:
            json.dump(ls_state, fp)
        return

    def load(self, filename):
        """ Reload important data into this class after a previous save.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data has been saved.

        """

        import json

        # Load file
        with open(filename, 'r') as fp:
            ls_state = json.load(fp)
        # Deserialize RBF object from dictionary
        self.m = ls_state['m']
        self.n = ls_state['n']
        self.loc_inds = ls_state['loc_inds']
        # Deserialize numpy.ndarray objects
        self.lb = np.array(ls_state['lb'])
        self.ub = np.array(ls_state['ub'])
        self.x_vals = np.array(ls_state['x_vals'])
        self.f_vals = np.array(ls_state['f_vals'])
        self.eps = np.array(ls_state['eps'])
        self.tr_center = np.array(ls_state['tr_center'])
        self.weights = np.array(ls_state['weights'])
        return




def test_sims_LinearSim():
    """ Test the linear sim function.

    Create an instance of sims.LinearSim and check that its output is correct.

    """

    from jax import config
    config.update("jax_enable_x64", True)
    import numpy as np
    from parmoo.simulations.simple import LinearSim

    dtype = np.dtype([(f"x{i+1}", "f8") for i in range(5)])
    ls = LinearSim(dtype, num_obj=3, num_r=5)
    x1 = {"x1": 0, "x2": 0, "x3": 0, "x4": 0, "x5": 0}
    x2 = {"x1": 1 / 3, "x2": 1 / 3, "x3": 1 / 3, "x4": 1 / 3, "x5": 1 / 3}
    x3 = {"x1": 2 / 3, "x2": 2 / 3, "x3": 2 / 3, "x4": 2 / 3, "x5": 2 / 3}
    assert (np.all(np.abs(ls(x1)[:5]) < 1.0e-8))
    assert (np.all(np.abs(ls(x2)[:5] - 1 / 3) < 1.0e-8))
    assert (np.all(np.abs(ls(x3)[:5] - 2 / 3) < 1.0e-8))
    assert (np.all(np.abs(ls(x1)[5:10] + 1 / 3) < 1.0e-8))
    assert (np.all(np.abs(ls(x2)[5:10]) < 1.0e-8))
    assert (np.all(np.abs(ls(x3)[5:10] - 1 / 3) < 1.0e-8))
    assert (np.all(np.abs(ls(x1)[10:] + 2 / 3) < 1.0e-8))
    assert (np.all(np.abs(ls(x2)[10:] + 1 / 3) < 1.0e-8))
    assert (np.all(np.abs(ls(x3)[10:]) < 1.0e-8))


def test_sims_QuadraticSim():
    """ Test the quadratic sim function.

    Create an instance of sims.QuadraticSim and check that its output is
    correct.

    """

    from jax import config
    config.update("jax_enable_x64", True)
    import numpy as np
    from parmoo.simulations.simple import QuadraticSim

    dtype = np.dtype([(f"x{i+1}", "f8") for i in range(5)])
    qs = QuadraticSim(dtype, num_obj=3, num_r=5)
    x1 = {"x1": 0, "x2": 0, "x3": 0, "x4": 0, "x5": 0}
    x2 = {"x1": 1 / 3, "x2": 1 / 3, "x3": 1 / 3, "x4": 1 / 3, "x5": 1 / 3}
    x3 = {"x1": 2 / 3, "x2": 2 / 3, "x3": 2 / 3, "x4": 2 / 3, "x5": 2 / 3}
    assert (np.abs(qs(x1)[0]) < 1.0e-8)
    assert (np.abs(qs(x2)[0]) > 1.0e-1)
    assert (np.abs(qs(x3)[0]) > 1.0e-1)
    assert (np.abs(qs(x1)[1]) > 1.0e-1)
    assert (np.abs(qs(x2)[1]) < 1.0e-8)
    assert (np.abs(qs(x3)[1]) > 1.0e-1)
    assert (np.abs(qs(x1)[2]) > 1.0e-1)
    assert (np.abs(qs(x2)[2]) > 1.0e-1)
    assert (np.abs(qs(x3)[2]) < 1.0e-8)


if __name__ == "__main__":
    test_sims_LinearSim()
    test_sims_QuadraticSim()



""" This module contains private methods for hosting and receiving callbacks
from an interactive dashboard. This module is intended only for developer use.

Note that some docstrings may be incomplete.

"""

import pandas as pd
import plotly.io as pio
from os import environ
from webbrowser import open_new
from warnings import warn
import logging
from dash import (
    Dash,
    Input,
    Output,
    html,
    dcc,
    exceptions,
    callback_context,
    no_update,
)
from .graph import (
    generate_scatter,
    generate_parallel,
    generate_radar,
)
from .utilities import (
    set_plot_name,
    set_database,
)


class Dash_App:
    """ A class for hosting the dashboard app. """

    def __init__(
        self,
        moop,
        plot_type,
        db,
        points,
        height,
        width,
        font,
        fontsize,
        background_color,
        screenshot,
        image_export_format,
        data_export_format,
        dev_mode,
        pop_up,
        port,
    ):
        """ Constructor for dashboard app. """

        logging.info('initializing dashboard')

        # ! STATE

        # * define independent state
        self.moop = moop
        self.plot_type = plot_type
        self.db = db
        self.points = points
        self.height = height
        self.width = width
        self.font = font
        self.fontsize = fontsize
        self.background_color = background_color
        self.screenshot = screenshot
        self.image_export_format = image_export_format
        self.data_export_format = data_export_format
        self.dev_mode = dev_mode
        self.pop_up = pop_up
        self.port = port

        # * define dependent state
        self.selection_indexes = []
        self.constraint_range = self.set_constraint_range(None)
        self.plot_name = set_plot_name(db=self.db)
        self.database = set_database(moop, db=self.db, points=self.points)
        self.graph = self.generate_graph()
        self.config = self.configure()

    # ! LAYOUT

        # * initialize app
        app = Dash(__name__)
        # * lay out app
        app.layout = html.Div(children=[
            dcc.Dropdown(
                id='plot_type_dropdown',
                options=['Scatterplot',
                         'Parallel Coordinates plot',
                         'Radar plot'],
                placeholder='Change plot type',
            ),
            dcc.Dropdown(
                id='database_dropdown',
                options=[
                    'Pareto Front',
                    'Objective Data'
                ],
                placeholder='Change dataset',
            ),
            html.Button(
                children='Export all data',
                id='download_dataset_button',
            ),
            dcc.Download(
                id='download_dataset_dcc',
            ),
            html.Button(
                children='Export selected data',
                id='download_selection_button',
            ),
            dcc.Download(
                id='download_selection_dcc',
            ),
            html.Button(
                children='Export image to working directory',
                id='download_image_button',
            ),
            dcc.Store(
                id='image'
            ),
            dcc.Checklist(
                id='constraint_checkboxes',
                options=[
                    {'label': 'show constraint-satisfying points',
                     'value': 'constraint_satisfying'},
                    {'label': 'show constraint-violating points',
                     'value': 'constraint_violating'},
                ],
                value=['constraint_satisfying'],
                inline=True,
            ),
            html.Br(),
            # * main plot
            dcc.Graph(
                id='parmoo_graph',
                figure=self.graph,
                config=self.config,
            ),
            html.Br(),
            html.Button(
                children='Show graph customization options',
                id='show_customization_options',
            ),
            html.Button(
                children='Hide graph customization options',
                id='hide_customization_options',
                style=dict(display='none'),
            ),
            html.Button(
                children='Show export options',
                id='show_export_options',
            ),
            html.Button(
                children='Hide export options',
                id='hide_export_options',
                style=dict(display='none'),
            ),
            dcc.Store(
                id='selection',
            ),
            dcc.Input(
                id='font_selection_input',
                placeholder='Set font',
                type='text',
                value='',
                debounce=True,
                style=dict(display='none'),
            ),
            dcc.Input(
                id='font_size_input',
                placeholder='Set font size',
                type='number',
                value='',
                min=1,
                max=100,
                debounce=True,
                style=dict(display='none'),
            ),
            dcc.Input(
                id='graph_width_input',
                placeholder='Set graph width',
                type='number',
                value='',
                min=10,
                step=1,
                style=dict(display='none'),
            ),
            dcc.Input(
                id='graph_height_input',
                placeholder='Set graph height',
                type='number',
                value='',
                min=10,
                step=1,
                style=dict(display='none'),
            ),
            dcc.Input(
                id='plot_name_input',
                placeholder='Set plot name',
                type='text',
                value='',
                debounce=True,
                style=dict(display='none'),
            ),
            dcc.Dropdown(
                id='background_color_dropdown',
                options=[
                    'White',
                    'Grey',
                    'Black',
                    'Transparent',
                ],
                placeholder='Set background color',
                style=dict(display='none'),
            ),
            dcc.Dropdown(
                id='image_export_format_dropdown',
                options=[
                    'PNG',
                    'JPEG',
                    'PDF',
                    'SVG',
                    'EPS',
                    'HTML',
                    'WebP'
                ],
                placeholder='Set image export format',
                style=dict(display='none'),
            ),
            dcc.Store(
                id='image_export_format_store',
                storage_type='local',
            ),
            dcc.Dropdown(
                id='data_export_format_dropdown',
                options=[
                    'CSV',
                    'JSON',
                ],
                placeholder='Set data export format',
                style=dict(display='none'),
            ),
            dcc.Store(
                id='data_export_format_store',
                storage_type='local',
            )
        ])

    # ! CALLBACKS

        # * show customization options
        @app.callback(
            Output(
                component_id='show_customization_options',
                component_property='style',),
            Output(
                component_id='hide_customization_options',
                component_property='style',),
            Output(
                component_id='font_selection_input',
                component_property='style',),
            Output(
                component_id='font_size_input',
                component_property='style',),
            Output(
                component_id='graph_width_input',
                component_property='style',),
            Output(
                component_id='graph_height_input',
                component_property='style',),
            Output(
                component_id='plot_name_input',
                component_property='style',),
            Output(
                component_id='background_color_dropdown',
                component_property='style',),
            Input(
                component_id='show_customization_options',
                component_property='n_clicks',),
            Input(
                component_id='hide_customization_options',
                component_property='n_clicks',),
            prevent_initial_call=True
        )
        def update_customization_components(
            s_clicks,
            h_clicks,
        ):
            """ Documentation incomplete. """

            triggered_id = callback_context.triggered[0]['prop_id']
            if triggered_id == 'show_customization_options.n_clicks':
                logging.info("'show_customization_options.n_clicks' triggered")
                return self.evaluate_customization_options('show', s_clicks)
            elif triggered_id == 'hide_customization_options.n_clicks':
                logging.info("'hide_customization_options.n_clicks' triggered")
                return self.evaluate_customization_options('hide', h_clicks)

        # * show export options
        @app.callback(
            Output(
                component_id='show_export_options',
                component_property='style',),
            Output(
                component_id='hide_export_options',
                component_property='style',),
            Output(
                component_id='image_export_format_dropdown',
                component_property='style',),
            Output(
                component_id='data_export_format_dropdown',
                component_property='style',),
            Input(
                component_id='show_export_options',
                component_property='n_clicks',),
            Input(
                component_id='hide_export_options',
                component_property='n_clicks',),
            prevent_initial_call=True
        )
        def update_export_components(
            s_clicks,
            h_clicks,
        ):
            """ Documentation incomplete. """

            triggered_id = callback_context.triggered[0]['prop_id']
            if triggered_id == 'show_export_options.n_clicks':
                logging.info("'show_export_options.n_clicks' triggered")
                return self.evaluate_export_options('show', s_clicks)
            elif triggered_id == 'hide_export_options.n_clicks':
                logging.info("'hide_export_options.n_clicks' triggered")
                return self.evaluate_export_options('hide', h_clicks)

        # * regenerate or update graph
        # TODO whenever possible, update instead of regenerating
        @app.callback(
            Output(
                component_id='parmoo_graph',
                component_property='figure',),
            # height - update
            Input(
                component_id='graph_height_input',
                component_property='value',),
            # width - update
            Input(
                component_id='graph_width_input',
                component_property='value',),
            # font - update
            Input(
                component_id='font_selection_input',
                component_property='value',),
            # font size - update
            Input(
                component_id='font_size_input',
                component_property='value',),
            # background color - update
            Input(
                component_id='background_color_dropdown',
                component_property='value',),
            # plot name - update
            Input(
                component_id='plot_name_input',
                component_property='value',),
            # plot type - regenerate
            Input(
                component_id='plot_type_dropdown',
                component_property='value',),
            # database - regenerate
            Input(
                component_id='database_dropdown',
                component_property='value',),
            # constraint show - regenerate
            Input(
                component_id='constraint_checkboxes',
                component_property='value'),
            prevent_initial_call=True
        )
        def update_graph(
            height_value,
            width_value,
            font_value,
            font_size_value,
            background_color_value,
            plot_name_value,
            plot_type_value,
            database_value,
            constraint_showr_value,
        ):
            """ Documentation incomplete. """

            triggered_id = callback_context.triggered[0]['prop_id']
            if 'graph_height_input.value' == triggered_id:
                logging.info("'graph_height_input.value' triggered")
                return self.evaluate_height(height_value)
            elif 'graph_width_input.value' == triggered_id:
                logging.info("'graph_width_input.value' triggered")
                return self.evaluate_width(width_value)
            elif 'font_selection_input.value' == triggered_id:
                logging.info("'font_selection_input.value' triggered")
                return self.evaluate_font(font_value)
            elif 'font_size_input.value' == triggered_id:
                logging.info("'font_size_input.value' triggered")
                return self.evaluate_font_size(font_size_value)
            elif 'background_color_dropdown.value' == triggered_id:
                logging.info("'background_color_dropdown.value' triggered")
                return self.evaluate_background_color(background_color_value)
            elif 'plot_name_input.value' == triggered_id:
                logging.info("'plot_name_input.value' triggered")
                return self.evaluate_plot_name(plot_name_value)
            elif 'plot_type_dropdown.value' == triggered_id:
                logging.info("'plot_type_dropdown.value' triggered")
                return self.evaluate_plot_type(plot_type_value)
            elif 'database_dropdown.value' == triggered_id:
                logging.info("'database_dropdown.value' triggered")
                return self.evaluate_database(database_value)
            elif 'constraint_checkboxes.value' == triggered_id:
                logging.info("'constraint_checkboxes.value' triggered")
                return self.evaluate_constraint_showr(constraint_showr_value)

        # * download dataset
        @app.callback(
            Output(
                component_id='download_dataset_dcc',
                component_property='data'),
            Input(
                component_id='download_dataset_button',
                component_property='n_clicks'),
        )
        def download_dataset(n_clicks):
            """ Documentation incomplete. """

            logging.info("'download_dataset_button.n_clicks' triggered")
            return self.evaluate_dataset_download(n_clicks)

        # * create object holding selected data
        @app.callback(
            Output(
                component_id='selection',
                component_property='data'),
            Input(
                component_id='parmoo_graph',
                component_property='selectedData'),
            Input(
                component_id='parmoo_graph',
                component_property='restyleData'),
        )
        def store_selection(
            selectedData,
            restyleData,
        ):
            """ Documentation incomplete. """

            triggered_id = callback_context.triggered[0]['prop_id']
            if 'parmoo_graph.selectedData' == triggered_id:
                logging.info("'parmoo_graph.selectedData' triggered")
                self.evaluate_selected_data(selectedData, 'selectedData')
            elif 'parmoo_graph.restyleData' == triggered_id:
                logging.info("'parmoo_graph.restyleData' triggered")
                if self.plot_type == 'parallel':
                    self.evaluate_selected_data(restyleData, 'restyleData')

        # * download selection
        @app.callback(
            Output(
                component_id='download_selection_dcc',
                component_property='data'),
            Input(
                component_id='download_selection_button',
                component_property='n_clicks'),
        )
        def download_selection(n_clicks):
            """ Documentation incomplete. """

            logging.info("'download_selection_button.n_clicks' triggered")
            return self.evaluate_selection_download(n_clicks)

        # * update image export format
        @app.callback(
            Output(
                component_id='image_export_format_store',
                component_property='data'),
            Input(
                component_id='image_export_format_dropdown',
                component_property='value'),
        )
        def update_image_export_format(value):
            """ Documentation incomplete. """

            logging.info("'image_export_format_dropdown.value' triggered")
            self.evaluate_image_export_format(value)

        # * update data export format
        @app.callback(
            Output(
                component_id='data_export_format_store',
                component_property='data'),
            Input(
                component_id='data_export_format_dropdown',
                component_property='value'),
        )
        def update_data_export_format(value):
            """ Documentation incomplete. """

            logging.info("'data_export_format_dropdown.value' triggered")
            self.evaluate_data_export_format(value)

        # * export image
        @app.callback(
            Output(
                component_id='image',
                component_property='data'),
            Input(
                component_id='download_image_button',
                component_property='n_clicks'),
        )
        def download_image(n_clicks):
            """ Documentation incomplete. """

            logging.info("'download_image_button.n_clicks' triggered")
            return self.evaluate_image_download(n_clicks)

    # ! EXECUTION

        logging.info('initialized dashboard')

        # * pop_up
        if pop_up:
            if not environ.get("WERKZEUG_RUN_MAIN"):
                open_new(port)

        # * run application
        logging.info('opening dashboard in browser. this might take a while')
        if dev_mode:
            app.run(
                debug=True,
                dev_tools_hot_reload=True,
            )
        elif not dev_mode:
            app.run(
                debug=False,
                dev_tools_hot_reload=False,
            )

    # ! INITIALIZATION HELPERS

    def generate_graph(self):
        """ Documentation incomplete. """

        if self.plot_type == 'scatter':
            self.graph = generate_scatter(
                moop=self.moop,
                db=self.db,
                points=self.points,
            )
        elif self.plot_type == 'parallel':
            self.graph = generate_parallel(
                moop=self.moop,
                db=self.db,
                points=self.points,
            )
        elif self.plot_type == 'radar':
            self.graph = generate_radar(
                moop=self.moop,
                db=self.db,
                points=self.points,
            )
        else:
            warn("invalid plot_type")

        self.graph = self.update_height()
        self.graph = self.update_width()
        self.graph = self.update_font()
        self.graph = self.update_font_size()
        self.graph = self.update_plot_name()
        self.graph = self.update_background_color()

        return self.graph

    def configure(self):
        """ Documentation incomplete. """

        if self.height != 'auto' and self.width != 'auto':
            self.config = {
                'displaylogo': False,
                'displayModeBar': True,
                'toImageButtonOptions': {
                    'format': self.screenshot,
                    'filename': str(self.plot_name),
                    'height': int(self.height),
                    'width': int(self.width),
                    'scale': 1
                }
            }
        else:
            self.config = {
                'displaylogo': False,
                'displayModeBar': True,
                'toImageButtonOptions': {
                    'format': self.screenshot,
                    'filename': str(self.plot_name),
                }
            }

        return self.config

    # ! UPDATE HELPERS

    # * functionality of select height input
    def update_height(self):
        """ Documentation incomplete. """

        if self.height != 'auto':
            self.graph.update_layout(
                height=int(self.height))
        return self.graph

    # * functionality of select width input
    def update_width(self):
        """ Documentation incomplete. """

        if self.width != 'auto':
            self.graph.update_layout(
                width=int(self.width))
        return self.graph

    # * functionality of select font input
    def update_font(self):
        """ Documentation incomplete. """

        if self.font != 'auto':
            self.graph.update_layout(
                font=dict(
                    family=self.font
                )
            )
        return self.graph

    # * functionality of select font size input
    def update_font_size(self):
        """ Documentation incomplete. """

        if self.fontsize != 'auto':
            self.graph.update_layout(
                font=dict(
                    size=int(self.fontsize)
                )
            )
        return self.graph

    # * functionality of plot name input
    def update_plot_name(self):
        """ Documentation incomplete. """

        self.graph.update_layout(
            title_text=self.plot_name
        )
        return self.graph

    # * functionality of background color dropdown
    def update_background_color(self):
        """ Documentation incomplete. """

        if self.background_color != 'auto':
            if self.plot_type == 'scatter':
                self.graph.update_layout(
                    plot_bgcolor=self.background_color,
                    paper_bgcolor=self.background_color,
                )
            elif self.plot_type == 'parallel':
                self.graph.update_layout(
                    paper_bgcolor=self.background_color,
                )
            elif self.plot_type == 'radar':
                self.graph.update_polars(
                    bgcolor=self.background_color,
                )
                self.graph.update_layout(
                    paper_bgcolor=self.background_color,
                )
        return self.graph

    # * functionality of plot type dropdown
    def update_plot_type(self):
        """ Documentation incomplete. """

        return self.generate_graph()

    # * functionality of database dropdown
    def update_database(self):
        """ Documentation incomplete. """

        self.database = set_database(
            moop=self.moop,
            db=self.db,
            points=self.points
        )
        if self.plot_name == 'Pareto Front':
            self.plot_name = set_plot_name(db=self.db)
        elif self.plot_name == 'Objective Data':
            self.plot_name = set_plot_name(db=self.db)
        return self.generate_graph()

    # ! CALLBACK HELPERS

    def evaluate_height(self, height_value):
        """ Documentation incomplete. """

        if height_value is not None:
            self.height = height_value
            return self.update_height()
        else:
            return self.graph

    def evaluate_width(self, width_value):
        """ Documentation incomplete. """

        if width_value is not None:
            self.width = width_value
            return self.update_width()
        else:
            return self.graph

    def evaluate_font(self, font_value):
        """ Documentation incomplete. """

        if font_value != '':
            self.font = font_value
            self.graph = self.update_font()
            return self.graph
        else:
            return self.graph

    def evaluate_font_size(self, font_size_value):
        """ Documentation incomplete. """

        self.fontsize = font_size_value
        self.graph = self.update_font_size()
        return self.graph

    def evaluate_background_color(self, background_color_value):
        """ Documentation incomplete. """

        if background_color_value == 'Transparent':
            self.background_color = 'rgb(0,0,0,0)'
        else:
            self.background_color = background_color_value
        return self.update_background_color()

    def evaluate_plot_name(self, plot_name_value):
        """ Documentation incomplete. """

        if plot_name_value != '':
            self.plot_name = plot_name_value
            self.graph = self.update_plot_name()
            return self.graph
        else:
            return self.graph

    def evaluate_plot_type(self, plot_type_value):
        """ Documentation incomplete. """

        if plot_type_value == 'Scatterplot':
            self.plot_type = 'scatter'
        elif plot_type_value == 'Parallel Coordinates plot':
            self.plot_type = 'parallel'
        elif plot_type_value == 'Radar plot':
            self.plot_type = 'radar'
        return self.update_plot_type()

    def evaluate_database(self, database_value):
        """ Documentation incomplete. """

        if database_value == 'Pareto Front':
            self.db = 'pf'
        elif database_value == 'Objective Data':
            self.db = 'obj'
        return self.update_database()

    def evaluate_dataset_download(self, n_clicks):
        """ Documentation incomplete. """

        if n_clicks is None:
            raise exceptions.PreventUpdate
        else:
            self.database.index.name = 'index'
            if self.data_export_format == 'CSV':
                return dict(
                    filename=str(self.plot_name) + ".csv",
                    content=self.database.to_csv(),
                )
            else:
                return dict(
                    filename=str(self.plot_name) + ".json",
                    content=self.database.to_json(),
                )

    def evaluate_selected_data(self, data, type):
        """ Documentation incomplete. """

        if data is None:
            raise exceptions.PreventUpdate
        else:
            self.selection_indexes = []
            if type == 'selectedData':
                selectedData = data
                pointskey = selectedData['points']
                for index in range(len(pointskey)):
                    level1 = pointskey[index]
                    point_index = level1['pointIndex']
                    self.selection_indexes.append(point_index)
            elif type == 'restyleData':
                self.update_constraint_range(data)
                objectives = self.moop.getObjectiveType().names
                for i, row in self.database.iterrows():
                    row_selected = True
                    for objective in objectives:
                        if row_selected:
                            row_obj_value = row[objective]
                            location = objectives.index(objective)
                            entry_dict = self.constraint_range[location]
                            ranges = entry_dict[list(entry_dict.keys())[0]]
                            if ranges is not None:
                                if row_selected:
                                    for range1 in ranges:
                                        try:
                                            for rang in range1:
                                                for ran in rang:
                                                    pass
                                            row_selected_yet = False
                                            for rang in range1:
                                                if not row_selected_yet:
                                                    if (row_obj_value >=
                                                       rang[0] and
                                                       row_obj_value <=
                                                       rang[1]):
                                                        row_selected_yet = True
                                                    else:
                                                        row_selected_yet = \
                                                            False
                                            row_selected = row_selected_yet
                                        except BaseException:
                                            if (row_obj_value >= range1[0] and
                                               row_obj_value <= range1[1]):
                                                row_selected = True
                                            else:
                                                row_selected = False
                    if row_selected:
                        self.selection_indexes.append(i)

    def set_constraint_range(self, restyleData):
        """ Documentation incomplete. """

        if restyleData is None:
            objectives = self.moop.getObjectiveType().names
            self.constraint_range = [None] * len(objectives)
            count = 0
            for objective in objectives:
                self.constraint_range[count] = dict({objective: None})
                count += 1
            return self.constraint_range
        else:
            return self.update_constraint_range(self, restyleData)

    def update_constraint_range(self, restyleData):
        """ Documentation incomplete. """

        key_list = restyleData[0]
        for key in key_list:
            location = int(key[11])
            entry_dict = self.constraint_range[location]
            entry_dict[list(entry_dict.keys())[0]] = key_list[key]
        return self.constraint_range

    def evaluate_selection_download(self, n_clicks):
        """ Documentation incomplete. """

        if n_clicks is None:
            raise exceptions.PreventUpdate
        else:
            selection_db = self.database.iloc[:0, :].copy()
            for i in self.selection_indexes:
                selection_db = pd.concat(
                    [selection_db,
                        self.database.iloc[[i]]]
                )
                selection_db.index.name = 'index'
            selection_db.drop_duplicates(inplace=True)
            selection_db.sort_index(inplace=True)
            if self.data_export_format == 'CSV':
                return dict(
                    filename="selected_data.csv",
                    content=selection_db.to_csv(),
                )
            else:
                return dict(
                    filename="selected_data.json",
                    content=selection_db.to_json(),
                )

    def evaluate_image_export_format(self, image_export_format_value):
        """ Documentation incomplete. """

        if image_export_format_value is not None:
            self.image_export_format = image_export_format_value

    def evaluate_data_export_format(self, data_export_format_value):
        """ Documentation incomplete. """

        if data_export_format_value is not None:
            self.data_export_format = data_export_format_value

    def evaluate_image_download(self, n_clicks):
        """ Documentation incomplete. """

        if n_clicks is None:
            raise exceptions.PreventUpdate
        else:
            file_name = str(self.plot_name)
            file_name += '.' + str(self.image_export_format)
            if (str(self.image_export_format) == 'HTML' or
               str(self.image_export_format) == 'html'):
                con_tent = pio.write_html(
                        fig=self.graph,
                        file=str(file_name)
                )
            else:
                con_tent = pio.write_image(
                        fig=self.graph,
                        file=str(file_name)
                )

            return dict(
                filename=file_name,
                content=con_tent,
            )

    def evaluate_customization_options(self, action, n_clicks):
        """ Documentation incomplete. """

        if n_clicks is None:
            raise exceptions.PreventUpdate
        else:
            showr = dict()
            hider = dict(display='none')
            if action == 'show':
                return hider, showr, showr, showr, showr, showr, showr, showr
            else:
                return showr, hider, hider, hider, hider, hider, hider, hider

    def evaluate_export_options(self, action, n_clicks):
        """ Documentation incomplete. """

        if n_clicks is None:
            return no_update, no_update
        else:
            showr = dict()
            hider = dict(display='none')
            if action == 'show':
                return hider, showr, showr, showr
            else:
                return showr, hider, hider, hider

    def evaluate_constraint_showr(self, value):
        """ Documentation incomplete. """

        """ Evaluate constraint toggles and update graph accordingly.

        Args:
            value (string): A string representing the state of the constraint
                toggles:
                 * 'constraint_satisfying' - 'Show constraint-satisfying
                   points' is the only toggle selected. Update graph to show
                   only points that satisfy every constraint.
                 * 'constraint_violating' - 'Show constraint-violating points'
                   is the only toggle selected. Update graph to show only
                   points that violate any constraint.
                 * 'all' - Both constraint toggles are selected. Update graph
                   to include all points.
                 * 'none' - No constraint toggles are selected. Update graph
                   to include no points.

        Returns:
            (plotly.graph_objects.Figure): A graph containing the points
            selected by the constraint toggles.

        """

        if value is None:
            raise exceptions.PreventUpdate
        else:
            if value == ['constraint_satisfying']:
                self.points = 'constraint_satisfying'
            elif value == ['constraint_violating']:
                self.points = 'constraint_violating'
            elif value == ['constraint_satisfying', 'constraint_violating']:
                self.points = 'all'
            elif value == ['constraint_violating', 'constraint_satisfying']:
                self.points = 'all'
            else:
                self.points = 'none'
        self.database = set_database(
            moop=self.moop,
            db=self.db,
            points=self.points,
        )
        return self.generate_graph()



def test_SingleSimObjective():
    """ Test the SingleSimObjective() function.

    Initialize an objective then evaluate.

    """

    from parmoo.objectives import SingleSimObjective, SingleSimGradient
    import numpy as np

    # Create named dtypes
    xtype = [("x1", "f8"), ("x2", "f8"), ("x3", "f8")]
    stype = [("sim1", "f8"), ("sim2", "f8", 2)]
    # Create input vectors
    x = np.zeros(1, dtype=xtype)[0]
    sx = np.zeros(1, dtype=stype)[0]
    sx['sim1'] = 2.0
    sx['sim2'][1] = 3.0
    # Create the objective and gradient functions
    obj_func1 = SingleSimObjective(xtype, stype, 'sim1', goal='max')
    grad_func1 = SingleSimGradient(xtype, stype, 'sim1', goal='max')
    obj_func2 = SingleSimObjective(xtype, stype, ('sim2', 1), goal='min')
    grad_func2 = SingleSimGradient(xtype, stype, ('sim2', 1), goal='min')
    # Test function evaluations
    assert (obj_func1(x, sx) == -2.0)
    assert (obj_func2(x, sx) == 3.0)
    # Test dx and ds evaluation
    xkeys = ["x1", "x2", "x3"]
    dx, ds = grad_func1(x, sx)
    assert (all([dx[i] == 0.0 for i in xkeys]))
    assert (ds["sim1"] == -1.0)
    assert (np.all(ds["sim2"] == 0.0))
    dx, ds = grad_func2(x, sx)
    assert (all([dx[i] == 0.0 for i in xkeys]))
    assert (ds["sim1"] == 0.0)
    assert (np.all(ds["sim2"] == np.eye(2)[1]))


def test_SumOfSimSquaresObjective():
    """ Test the SumOfSimSquaresObjective() function.

    Initialize an objective then evaluate.

    """

    from parmoo.objectives import SumOfSimSquaresObjective
    from parmoo.objectives import SumOfSimSquaresGradient
    import numpy as np

    # Create named dtypes
    xtype = [("x1", "f8"), ("x2", "f8"), ("x3", "f8"), ("x4", "f8")]
    stype = [("sim1", "f8"), ("sim2", "f8", 4)]
    # Create input vectors
    x = np.ones(1, dtype=xtype)[0]
    sx = np.ones(1, dtype=stype)[0]
    sx['sim1'] = 2.0
    sx['sim2'][:] = 2.0
    # Create the objective and gradient functions
    sim_list = ['sim1', ('sim2', 0), ('sim2', 1), ('sim2', 2)]
    obj_func = SumOfSimSquaresObjective(xtype, stype, sim_list, goal="max")
    grad_func = SumOfSimSquaresGradient(xtype, stype, sim_list, goal="max")
    # Test function evaluation
    assert (np.abs(obj_func(x, sx) + 16.0) < 1.0e-8)
    # Test dx and ds evaluation
    xkeys = ["x1", "x2", "x3", "x4"]
    dx, ds = grad_func(x, sx)
    assert (all([dx[i] == 0.0 for i in xkeys]))
    assert (ds["sim1"] == -4.0)
    assert (np.all(ds["sim2"] == (np.eye(4)[3] - np.ones(4)) * 4.0))


def test_SumOfSimsObjective():
    """ Test the SumOfSimsObjective() function.

    Initialize an objective then evaluate.

    """

    from parmoo.objectives import SumOfSimsObjective, SumOfSimsGradient
    import numpy as np

    # Create named dtypes
    xtype = [("x1", "f8"), ("x2", "f8"), ("x3", "f8"), ("x4", "f8")]
    stype = [("sim1", "f8"), ("sim2", "f8", 4)]
    # Create input vectors
    x = np.ones(1, dtype=xtype)[0]
    sx = np.ones(1, dtype=stype)[0]
    sx['sim1'] = 2.0
    sx['sim2'][:] = -2.0
    # Create the objective functions
    sim_list = ['sim1', ('sim2', 0), ('sim2', 1), ('sim2', 2)]
    obj_func1 = SumOfSimsObjective(xtype, stype, sim_list, goal="max")
    grad_func1 = SumOfSimsGradient(xtype, stype, sim_list, goal="max")
    obj_func2 = SumOfSimsObjective(xtype, stype, sim_list,
                                   goal='min', absolute=True)
    grad_func2 = SumOfSimsGradient(xtype, stype, sim_list,
                                   goal='min', absolute=True)
    # Test function evaluation
    assert (np.abs(obj_func1(x, sx) - 4.0) < 1.0e-8)
    assert (np.abs(obj_func2(x, sx) - 8.0) < 1.0e-8)
    # Test dx and ds evaluation
    xkeys = ["x1", "x2", "x3", "x4"]
    dx, ds = grad_func1(x, sx)
    assert (all([dx[i] == 0.0 for i in xkeys]))
    assert (ds["sim1"] == -1.0)
    assert (np.all(ds["sim2"] == (np.eye(4)[3] - np.ones(4))))
    dx, ds = grad_func2(x, sx)
    assert (all([dx[i] == 0.0 for i in xkeys]))
    assert (ds["sim1"] == 1.0)
    assert (np.all(ds["sim2"] == (np.eye(4)[3] - np.ones(4))))


if __name__ == "__main__":
    test_SingleSimObjective()
    test_SumOfSimSquaresObjective()
    test_SumOfSimsObjective()




import numpy as np
from parmoo import MOOP
from parmoo.searches import LatinHypercube
from parmoo.surrogates import GaussRBF
from parmoo.optimizers import GlobalSurrogate_PS

my_moop = MOOP(GlobalSurrogate_PS)

# Define a simulation to use below
def sim_func(x):
    if x["MyCat"] == 0:
        return np.array([(x["MyDes"]) ** 2, (x["MyDes"] - 1.0) ** 2])
    else:
        return np.array([99.9, 99.9])

# Add a design variable, simulation, objective, and constraint.
# Note the 'name' keys for each
my_moop.addDesign({'name': "MyDes",
                   'des_type': "continuous",
                   'lb': 0.0, 'ub': 1.0})
my_moop.addDesign({'name': "MyCat",
                   'des_type': "categorical",
                   'levels': 2})

my_moop.addSimulation({'name': "MySim",
                       'm': 2,
                       'sim_func': sim_func,
                       'search': LatinHypercube,
                       'surrogate': GaussRBF,
                       'hyperparams': {'search_budget': 20}})

my_moop.addObjective({'name': "MyObj",
                      'obj_func': lambda x, s: sum(s["MySim"])})

my_moop.addConstraint({'name': "MyCon",
                       'constraint': lambda x, s: 0.1 - x["MyDes"]})

# Extract numpy dtypes for all of this MOOP's inputs/outputs
des_dtype = my_moop.getDesignType()
obj_dtype = my_moop.getObjectiveType()
sim_dtype = my_moop.getSimulationType()

# Display the dtypes as strings
print("Design variable type:   " + str(des_dtype))
print("Simulation output type: " + str(sim_dtype))
print("Objective type:         " + str(obj_dtype))




""" Use ParMOO to solve the DTLZ3 problem, treating DTLZ3 as a simulation.

Uses named variables, the dtlz3_sim simulation function, and the
SingleSimObjective objective functions to define the problem.

Also activates ParMOO's checkpointing feature, in order to test checkpointing.

"""

import numpy as np
import os
from parmoo import MOOP
from parmoo.optimizers import GlobalSurrogate_PS
from parmoo.surrogates import GaussRBF
from parmoo.acquisitions import FixedWeights
from parmoo.searches import LatinHypercube
from parmoo.simulations.dtlz import dtlz3_sim as sim_func
from parmoo.objectives import SingleSimObjective as obj_func

# Set the problem dimensions
NUM_DES = 3
NUM_OBJ = 3

# Create a MOOP
moop = MOOP(GlobalSurrogate_PS)

# Add NUM_DES continuous design variables
for i in range(NUM_DES):
    moop.addDesign({'name': f"x{i+1}", 'ub': 1.0, 'lb': 0.0,
                    'des_type': "continuous", 'des_tol': 1.0e-8})

# Add the simulation
moop.addSimulation({'name': "DTLZ3",
                    'm': NUM_OBJ,
                    'sim_func': sim_func(moop.getDesignType(),
                                         num_obj=NUM_OBJ,
                                         offset=0.6),
                    'search': LatinHypercube,
                    'surrogate': GaussRBF,
                    'hyperparams': {}})

# Add NUM_OBJ objective functions
for i in range(NUM_OBJ):
    moop.addObjective({'name': f"f{i+1}",
                       'obj_func': obj_func(moop.getDesignType(),
                                            moop.getSimulationType(),
                                            ("DTLZ3", i), goal="min")})

# Define 2 constraints to nudge the solver in the right direction

def min_constraint(x, sx):
    """ x[NUM_OBJ-1:NUM_DES] >= 0.5 """

    fx = 0.0
    for i in range(NUM_OBJ - 1, NUM_DES):
        fx += (0.5 - x[f"x{i+1}"])
    return fx

def max_constraint(x, sx):
    """ x[NUM_OBJ-1:NUM_DES] <= 0.7 """

    fx = 0.0
    for i in range(NUM_OBJ - 1, NUM_DES):
        fx += (x[f"x{i+1}"] - 0.7)
    return fx

# Add 2 constraints to the problem
moop.addConstraint({'name': "Lower Bounds", 'constraint': min_constraint})
moop.addConstraint({'name': "Upper Bounds", 'constraint': max_constraint})

# Add NUM_OBJ acquisition functions
for i in range(NUM_OBJ):
    moop.addAcquisition({'acquisition': FixedWeights,
                         'hyperparams': {'weights': np.eye(NUM_OBJ)[i]}})


# Solve the problem with 5 iterations + checkpointing
moop.setCheckpoint(True)
moop.solve(5)

# Check that 115 simulations were evaluated and solutions are feasible
assert(moop.getObjectiveData().shape[0] == (100 + NUM_OBJ*5))
assert(moop.getSimulationData()['DTLZ3'].shape[0] == (100 + NUM_OBJ*5))
assert(moop.getPF().shape[0] > 0)

# Clean up test directory (remove checkpoint files)
os.remove("parmoo.moop")
os.remove("parmoo.simdb.json")
os.remove("parmoo.surrogate.1")
